# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from functools import reduce
from operator import mul
from typing import Dict, List

import torch
from torch.fx.node import Node, _get_qualified_name, map_arg

from .bridge import to_torch_spmd
from .device_mesh import device_mesh_world_size

DTYPE_BYTE = {
    torch.float32: 4,
    torch.float16: 2,
    torch.int32: 4,
    torch.int64: 8,
    torch.uint8: 1,
    torch.bool: 0.125
}


def tensor_mem_size(shape_info, mem_ratio):
    mem_size = 0
    if isinstance(shape_info, tuple):
        for shape_info_, mem_ratio_ in zip(shape_info, mem_ratio):
            if 'shape' in shape_info_:
                if len(shape_info_['shape']) == 0:
                    mem_size += DTYPE_BYTE[shape_info_['dtype']]
                else:
                    mem_size += reduce(
                        mul, shape_info_['shape']) * DTYPE_BYTE[shape_info_['dtype']] * mem_ratio_
    else:
        if 'shape' in shape_info:
            if len(shape_info['shape']) == 0:
                mem_size += DTYPE_BYTE[shape_info['dtype']]
            else:
                mem_size = reduce(
                    mul, shape_info['shape']) * DTYPE_BYTE[shape_info['dtype']] * mem_ratio
    return mem_size


def _get_input_strategy(fx_graph, opt_strategy):
    input_strategy = {}

    input_name = []
    for node in fx_graph.graph.nodes:
        if node.op == 'placeholder':
            input_name.append(node.name)

    for node in reversed(fx_graph.graph.nodes):
        if node.op == "call_function":
            op_name = _get_qualified_name(node.target)
            if op_name != "_operator.getitem":
                unique_key = node.name
                if unique_key in opt_strategy:
                    invars = opt_strategy[unique_key]['node'].invars
                    for idx, var in enumerate(invars):
                        if var.name in input_name:
                            input_strategy[var.name] = [
                                to_torch_spmd(i)
                                for i in opt_strategy[unique_key]['strategy'].in_strtg_group[idx]
                            ]

    return input_strategy


def mem_anaylize(fx_graph: torch.fx.GraphModule, tensor_shape_info, opt_strategy):

    same_mem_funcs = [
        "torch.ops.aten.t", "torch.ops.aten.view", "torch.ops.aten._unsafe_view",
        "torch.ops.aten.slice", "torch.ops.aten.transpose", "torch.ops.aten.squeeze",
        "torch.ops.aten.unsqueeze", "torch.ops.aten.permute", "torch.ops.aten.slice",
        "_operator.getitem", "torch.ops.aten.expand"
    ]

    # Run through reverse nodes and record the first instance of a use
    # of a given node. This represents the *last* use of the node in the
    # execution order of the program, which we will use to free unused
    # values
    node_to_last_use: Dict[Node, Node] = {}
    user_to_last_uses: Dict[Node, List[Node]] = {}

    def register_last_uses(n: Node, user: Node):
        if n not in node_to_last_use:
            node_to_last_use[n] = user
            user_to_last_uses.setdefault(user, []).append(n)

    for node in reversed(fx_graph.graph.nodes):
        map_arg(node.args, lambda n: register_last_uses(n, node))
        map_arg(node.kwargs, lambda n: register_last_uses(n, node))

    persistent_mem, activation_mem = {}, {}
    mem_ratio = {}
    same_memory_tensor_info = {}

    total_memory, peak_memory = 0, 0

    f = open("mem_info.txt", "w")

    world_size = device_mesh_world_size()

    input_strategy = _get_input_strategy(fx_graph, opt_strategy)

    for node in fx_graph.graph.nodes:
        # input and parametes will not release during inference
        if node.op == 'placeholder':

            mem_ratio[node.name] = 1
            if node.name in input_strategy:
                for place in input_strategy[node.name]:
                    if place.is_shard():
                        mem_ratio[node.name] = 1 / world_size

            persistent_mem[node.name] = tensor_mem_size(tensor_shape_info[node.name],
                                                        mem_ratio[node.name])
            total_memory += persistent_mem[node.name]
            if peak_memory < total_memory:
                peak_memory = total_memory

        if node.op == 'call_function':
            if _get_qualified_name(node.target) in same_mem_funcs:
                if node.args[0].name not in persistent_mem:
                    if node.args[0] not in same_memory_tensor_info:
                        same_memory_tensor_info[node.args[0].name] = [node.args[0].name]
                    same_memory_tensor_info[node.args[0].name].append(node.name)
            else:
                mem_ratio[node.name] = [1]
                if node.name in opt_strategy:
                    out_strategy = [[to_torch_spmd(i) for i in ii]
                                    for ii in opt_strategy[node.name]['strategy'].out_strtg_group]
                    mem_ratio[node.name] = [1] * len(out_strategy)
                    for idx, out_stgy in enumerate(out_strategy):
                        for place in out_stgy:
                            if place.is_shard():
                                mem_ratio[node.name][idx] = 1 / world_size

                if not isinstance(tensor_shape_info[node.name], tuple):
                    mem_ratio[node.name] = mem_ratio[node.name][0]

                activation_mem[node.name] = tensor_mem_size(tensor_shape_info[node.name],
                                                            mem_ratio[node.name])
                total_memory += activation_mem[node.name]
                if peak_memory < total_memory:
                    peak_memory = total_memory

        # release used tensor
        for to_delete in user_to_last_uses.get(node, []):

            for value in same_memory_tensor_info.values():
                while to_delete.name in value:
                    value.remove(to_delete.name)

            new_same_memory_tensor_info = {}
            for k, v in same_memory_tensor_info.items():
                if len(v) > 0:
                    new_same_memory_tensor_info[k] = v
                else:
                    if k in activation_mem:
                        total_memory -= tensor_mem_size(tensor_shape_info[k], mem_ratio[k])
                        del activation_mem[k]
            same_memory_tensor_info = new_same_memory_tensor_info

            if to_delete.name in activation_mem:
                if to_delete.name not in same_memory_tensor_info:
                    total_memory -= tensor_mem_size(tensor_shape_info[to_delete.name],
                                                    mem_ratio[to_delete.name])
                    del activation_mem[to_delete.name]

        gb_ratio = 1024 * 1024 * 1024
        f.write(
            f"Node | Total | Peak Memory: {node} | {total_memory / gb_ratio:.4f} GB | {peak_memory / gb_ratio:.4f} GB\n"
        )

    f.close()



# Register sharding propagation rules for some ops. These rules are missed in community pytorch.
# refer to https://github.com/pytorch/pytorch/tree/main/torch/distributed/_tensor/ops.

import copy
import functools
import math
import operator
from functools import reduce
from typing import cast

import torch
from torch._meta_registrations import calc_conv_nd_return_shape
from torch.distributed._tensor.op_schema import OpSchema, OutputSharding
from torch.distributed._tensor.ops.common_rules import pointwise_rule
from torch.distributed._tensor.ops.utils import register_prop_rule
from torch.distributed._tensor.placement_types import (DTensorSpec, Replicate, Shard, _Partial)

aten = torch.ops.aten

extra_pointwise_op = [
    aten.silu.default, aten.leaky_relu.default, aten.leaky_relu_backward.default, aten.elu.default,
    aten.elu_backward.default, aten.fill.Scalar
]

for op in extra_pointwise_op:
    register_prop_rule(op)(pointwise_rule)


# from torch/distributed/_spmd/experimental_ops.py on pytorch main branch
# copy it because it is not ready in released pytorch
@register_prop_rule(aten.native_layer_norm.default)  # pyre-ignore
def _prop_native_layer_norm(op_schema: OpSchema) -> OutputSharding:
    input, normalized_shape, weight, bias, eps = op_schema.args_schema
    assert isinstance(input, DTensorSpec)
    assert isinstance(normalized_shape, (tuple, list))
    if weight is not None:
        assert isinstance(weight, DTensorSpec)
        assert all(isinstance(p, Replicate) for p in weight.placements)
    if bias is not None:
        assert isinstance(bias, DTensorSpec)
        assert all(isinstance(p, Replicate) for p in bias.placements)
    # only the left-most (non-normalized) dimensions of the input can be sharded
    batch_ndim = len(input.shape) - len(normalized_shape)
    assert all(
        isinstance(p, Replicate) or (isinstance(p, Shard) and p.dim < batch_ndim, )
        for p in input.placements)
    stats_spec = DTensorSpec(
        mesh=input.mesh,
        placements=input.placements,
        ndim=input.ndim,
        shape=torch.Size(s for (i, s) in enumerate(input.shape)),
    )
    return OutputSharding(output_spec=(input, stats_spec, stats_spec))


# from torch/distributed/_spmd/experimental_ops.py on pytorch main branch
# copy it because it is not ready in released pytorch
@register_prop_rule(aten.native_layer_norm_backward.default)  # pyre-ignore
def _prop_native_layer_norm_backward(op_schema: OpSchema) -> OutputSharding:
    (
        grad,
        input,
        normalized_shape,
        result1,
        result2,
        weight,
        bias,
        grad_input_mask,
    ) = op_schema.args_schema
    assert isinstance(grad, DTensorSpec)
    assert isinstance(grad_input_mask, (list, tuple))
    if weight is not None:
        assert isinstance(weight, DTensorSpec)
        assert all(isinstance(s, Replicate) for s in weight.placements)
    if bias is not None:
        assert isinstance(bias, DTensorSpec)
        assert all(isinstance(s, Replicate) for s in bias.placements)

    if all(isinstance(s, Replicate) for s in grad.placements):
        weight_grad = (DTensorSpec(
            mesh=weight.mesh,
            placements=[Replicate()] * weight.mesh.ndim,
            shape=weight.shape,
        ) if weight else None)
        bias_grad = (DTensorSpec(
            mesh=bias.mesh,
            placements=[Replicate()] * bias.mesh.ndim,
            shape=bias.shape,
        ) if bias else None)
        return OutputSharding(
            # NOTE: type errors below are legit. This is because DTensor currently
            # doesn't support Optional return values. Need to be fixed in DTensor repo.
            output_spec=(
                grad if grad_input_mask[0] else None,
                weight_grad if grad_input_mask[1] else None,
                bias_grad if grad_input_mask[2] else None,
            ), )

    batch_ndim = len(input.shape) - len(normalized_shape)
    assert any(isinstance(s, Shard) and s.dim < batch_ndim
               for s in grad.placements), f"Got {grad.placements}"
    weight_grad = (DTensorSpec(
        mesh=weight.mesh,
        placements=[_Partial()] * weight.mesh.ndim,
        shape=weight.shape,
    ) if weight else None)
    bias_grad = (DTensorSpec(
        mesh=bias.mesh,
        placements=[_Partial()] * bias.mesh.ndim,
        shape=bias.shape,
    ) if bias else None)
    return OutputSharding(
        # NOTE: type errors below are legit. This is because DTensor currently
        # doesn't support Optional return values. Need to be fixed in DTensor repo.
        output_spec=(
            grad if grad_input_mask[0] else None,
            weight_grad if grad_input_mask[1] else None,
            bias_grad if grad_input_mask[2] else None,
        ), )


@register_prop_rule(aten.convolution.default)
def _prop_convolution_default(op_schema: OpSchema) -> OutputSharding:
    input, weight = op_schema.args_schema[0:2]
    stride, padding, dialation, transposed, output_padding, groups = op_schema.args_schema[-6:]
    assert isinstance(input, DTensorSpec)
    assert isinstance(weight, DTensorSpec)
    output_placements = [Replicate()] * input.mesh.ndim

    if len(output_padding) == 0 or all(v == 0 for v in output_padding):
        output_padding = None
    output_shape = calc_conv_nd_return_shape(input, weight, stride, padding, dialation, transposed,
                                             groups, output_padding)

    input_shape = input.shape
    for idx, s in enumerate(input.placements):
        # if shard on batch, height, witdh
        if isinstance(s, Shard) and s.dim in [0] + list(range(2, len(input_shape))):
            output_placements[idx] = Shard(dim=s.dim)

    for idx, s in enumerate(weight.placements):
        if isinstance(s, Shard) and s.dim == 0:
            output_placements[idx] = Shard(dim=1)
        s_input = input.placements[idx]
        if isinstance(s, Shard) and s.dim == 1 and isinstance(s_input, Shard) and s_input.dim == 1:
            output_placements[idx] = _Partial()

    output_spec = DTensorSpec(mesh=input.mesh,
                              placements=output_placements,
                              shape=torch.Size(output_shape))

    return OutputSharding(output_spec=output_spec)


@register_prop_rule(aten.convolution_backward.default)
def _prop_convolution_backward_default(op_schema: OpSchema) -> OutputSharding:
    (grad_output, input, weight, bias_sizes, stride, padding, dilation, transposed, output_padding,
     groups, output_mask) = op_schema.args_schema

    weight_placement = [Replicate()] * weight.mesh.ndim
    bias_placement = [Replicate()] * weight.mesh.ndim

    for idx, s in enumerate(input.placements):
        if isinstance(s, Shard) and s.dim == 0:
            weight_placement[idx] = _Partial()
            bias_placement[idx] = _Partial()
        else:
            weight_placement[idx] = weight.placements[idx]
            if isinstance(weight_placement[idx], Shard) and weight_placement[idx].dim == 0:
                bias_placement[idx] = Shard(dim=0)

    weight_grad = DTensorSpec(mesh=weight.mesh, placements=weight_placement, shape=weight.shape)

    bias_grad = DTensorSpec(mesh=weight.mesh,
                            placements=bias_placement,
                            shape=torch.Size(bias_sizes))

    return OutputSharding(output_spec=(
        input if output_mask[0] else None,
        weight_grad if output_mask[1] else None,
        bias_grad if output_mask[2] else None,
    ), )


@register_prop_rule(aten.native_batch_norm.default)
def _prop_batch_norm_default(op_schema: OpSchema) -> OutputSharding:
    input, weight, bias = op_schema.args_schema[0:3]
    running_mean, running_var = op_schema.args_schema[3:5]

    return OutputSharding(output_spec=(input, running_mean, running_var))


@register_prop_rule(aten.cudnn_batch_norm.default)
def _prop_batch_norm_default(op_schema: OpSchema) -> OutputSharding:
    input, weight, bias = op_schema.args_schema[0:3]
    running_mean, running_var = op_schema.args_schema[3:5]

    reserve_placement = [Replicate()] * weight.mesh.ndim

    reserve = DTensorSpec(mesh=weight.mesh, placements=reserve_placement, shape=torch.Size([0]))

    return OutputSharding(output_spec=(input, running_mean, running_var, reserve), )


@register_prop_rule(aten._native_batch_norm_legit_functional.default)
def _prop_batch_norm_legit_functional_default(op_schema: OpSchema) -> OutputSharding:

    input, weight, bias = op_schema.args_schema[0:3]
    running_mean, running_var = op_schema.args_schema[3:5]

    return OutputSharding(output_spec=(input, running_mean, running_var, running_mean,
                                       running_var), )


@register_prop_rule(aten.native_group_norm.default)
def _prop_native_group_norm(op_schema: OpSchema) -> OutputSharding:

    input, weight, bias, N, C, HxW, group, eps = op_schema.args_schema

    local_shape = input.local_shape
    flatten_inner_shape = functools.reduce((lambda x, y: x * y), local_shape[2:])

    op_schema.args_schema = input, weight, bias, local_shape[0], local_shape[
        1], flatten_inner_shape, group, eps

    assert isinstance(input, DTensorSpec)
    if weight is not None:
        assert isinstance(weight, DTensorSpec)
        assert all(isinstance(p, Replicate) for p in weight.placements)
    if bias is not None:
        assert isinstance(bias, DTensorSpec)
        assert all(isinstance(p, Replicate) for p in bias.placements)
    # only the batch (1st) dimensions of the input can be sharded
    batch_ndim = 1
    assert all(
        isinstance(p, Replicate) or (isinstance(p, Shard) and p.dim < batch_ndim, )
        for p in input.placements)

    stats_placement = [Replicate(), Replicate()]
    for idx, p in enumerate(input.placements):
        if isinstance(p, Shard) and p.dim < batch_ndim:
            stats_placement[idx] = copy.deepcopy(p)
    stats_spec = DTensorSpec(
        mesh=input.mesh,
        placements=stats_placement,
        ndim=input.mesh.ndim,
        shape=torch.Size((input.shape[0], group)),
    )
    return OutputSharding(output_spec=(input, stats_spec, stats_spec))


@register_prop_rule(aten.native_group_norm_backward.default)
def _prop_native_group_norm_backward(op_schema: OpSchema) -> OutputSharding:
    grad_output, input, mean, rstd, gamma, N, C, HxW, group, output_mask = op_schema.args_schema

    local_shape = input.local_shape

    op_schema.args_schema = grad_output, input, mean, rstd, gamma, local_shape[0], local_shape[
        1], reduce(operator.mul, local_shape[2:]), group, output_mask

    spec = Replicate()
    if isinstance(input.placements[0], Shard):
        spec = _Partial()
    d_gamma = DTensorSpec(mesh=gamma.mesh, placements=[spec] * gamma.mesh.ndim, shape=gamma.shape)

    d_bias = DTensorSpec(mesh=gamma.mesh, placements=[spec] * gamma.mesh.ndim, shape=gamma.shape)

    return OutputSharding(
        # NOTE: type errors below are legit. This is because DTensor currently
        # doesn't support Optional return values. Need to be fixed in DTensor repo.
        output_spec=(
            input if output_mask[0] else None,
            d_gamma if output_mask[1] else None,
            d_bias if output_mask[2] else None,
        ), )


@register_prop_rule(
    [aten.cudnn_batch_norm_backward.default, aten.native_batch_norm_backward.default])
def _prop_batch_norm_backward_default(op_schema: OpSchema) -> OutputSharding:
    input, grad_output, weight, running_mean, running_var = op_schema.args_schema[0:5]

    weight_placement = [Replicate()] * weight.mesh.ndim

    for idx, s in enumerate(weight.placements):
        grad_place = grad_output.placements[idx]
        if isinstance(s, Replicate) and isinstance(grad_place, Shard) and grad_place.dim == 0:
            weight_placement[idx] = _Partial()
        if isinstance(s, Shard):
            weight_placement[idx] = weight.placements[idx]

    weight_grad = DTensorSpec(mesh=weight.mesh, placements=weight_placement, shape=weight.shape)
    bias_grad = DTensorSpec(mesh=weight.mesh, placements=weight_placement, shape=weight.shape)

    return OutputSharding(
        # NOTE: type errors below are legit. This is because DTensor currently
        # doesn't support Optional return values. Need to be fixed in DTensor repo.
        output_spec=(
            input,
            weight_grad,
            bias_grad,
        ), )


@register_prop_rule(aten.max_pool2d_with_indices.default)
def _prop_max_pool2d_with_indices(op_schema: OpSchema) -> OutputSharding:
    input, kernel, stride = op_schema.args_schema[:3]
    if len(op_schema.args_schema) >= 4:
        padding = op_schema.args_schema[3]
    else:
        padding = [0, 0]

    output_shape = list(input.shape)

    for inner_dim in range(2):
        output_shape[2 + inner_dim] = int(
            math.floor(input.shape[2 + inner_dim] + 2 * padding[inner_dim] -
                       (kernel[inner_dim] - 1) - 1) / stride[inner_dim] + 1)

    output_spec = DTensorSpec(mesh=input.mesh,
                              placements=input.placements,
                              shape=torch.Size(output_shape))

    return OutputSharding(output_spec=(output_spec, input))


@register_prop_rule([
    aten.max_pool2d_with_indices_backward.default,
    aten.avg_pool3d_backward.default,
])
def _prop_max_pool2d_with_indices_backward(op_schema: OpSchema) -> OutputSharding:
    grad_output, input = op_schema.args_schema[0:2]

    return OutputSharding(output_spec=DTensorSpec(
        mesh=grad_output.mesh, placements=grad_output.placements, shape=input.shape))


@register_prop_rule(aten.constant_pad_nd.default)
def _prop_constant_pad_nd(op_schema: OpSchema) -> OutputSharding:
    input, pad, value = op_schema.args_schema

    input_sizes = input.shape
    l_inp = len(input_sizes)

    l_pad = len(pad) // 2
    l_diff = l_inp - l_pad

    new_shape = list(input_sizes[:l_diff])

    for i in range(l_pad):
        pad_idx = len(pad) - ((i + 1) * 2)
        new_dim = input_sizes[l_diff + i] + pad[pad_idx] + pad[pad_idx + 1]
        new_shape.append(new_dim)

    output_spec = DTensorSpec(mesh=input.mesh,
                              placements=input.placements,
                              shape=torch.Size(new_shape))

    return OutputSharding(output_spec=output_spec)


@register_prop_rule(aten.embedding.default)
def _prop_embedding(op_schema: OpSchema) -> OutputSharding:
    weight_spec, inp_spec = op_schema.args_spec
    if any(placement.is_shard(0) for placement in weight_spec.placements):
        raise NotImplementedError(
            "DTensor does not support row-wise sharded embedding operation yet!")

    out_shape = inp_spec.shape + weight_spec.shape[1:]

    if all(placement.is_replicate()
           for placement in weight_spec.placements) and inp_spec.placements == [Shard(0)]:
        # Embedding table is replicated, input ids are sharded along batch
        # dimension. Output lookups should match input sharding spec in this case.
        return OutputSharding(output_spec=DTensorSpec(
            mesh=inp_spec.mesh, placements=inp_spec.placements, shape=out_shape))

    if all(placement.is_replicate() for placement in inp_spec.placements):
        weight_dim_map = weight_spec.dim_map
        output_dim_map = inp_spec.dim_map
        output_dim_map.append(weight_dim_map[1])
        return OutputSharding(output_spec=DTensorSpec.from_dim_map(
            inp_spec.mesh, output_dim_map, [], shape=out_shape))

    return OutputSharding(
        output_spec=None,
        schema_suggestions=[
            OpSchema(
                func_schema=op_schema.func_schema,
                args_schema=(
                    weight_spec,
                    DTensorSpec(mesh=inp_spec.mesh,
                                placements=[Replicate()] * len(inp_spec.placements),
                                shape=inp_spec.shape,
                                ndim=inp_spec.ndim),
                ),
                kwargs_schema=op_schema.kwargs_schema,
            )
        ],
    )


# from torch/distributed/_spmd/experimental_ops.py on pytorch main branch
# copy it because it is not ready in released pytorch
@register_prop_rule(aten.select.int)
def _prop_select(op_schema: OpSchema) -> OutputSharding:

    tensor, dim = op_schema.args_schema[:2]
    assert isinstance(tensor, DTensorSpec)
    assert isinstance(dim, int)
    placements = tensor.placements
    assert all(not p.is_shard(dim)
               for p in placements), "DTensor does not support select on sharded dimension."

    # select will remove one dimension, decrement dim of Shard placements by 1
    # if they are larger than dim.
    new_placements = []
    for p in placements:
        # Using isinstance instead of is_shard so that mypy won't complain
        # about accessing dim attribute.
        if isinstance(p, Shard) and p.dim > dim:
            new_placements.append(Shard(p.dim - 1))
        else:
            new_placements.append(p)

    new_shape = tensor.shape[:dim] + tensor.shape[dim + 1:]

    return OutputSharding(output_spec=DTensorSpec(
        mesh=tensor.mesh, placements=new_placements, shape=new_shape, ndim=tensor.ndim - 1))


@register_prop_rule(aten.slice_backward.default)
def _prop_slice_backward(op_schema: OpSchema) -> OutputSharding:
    grad, input_sizes = op_schema.args_schema[0:2]

    return OutputSharding(output_spec=DTensorSpec(
        mesh=grad.mesh, placements=grad.placements, shape=input_sizes, ndim=grad.ndim))


@register_prop_rule(  # pyre-ignore
    [
        aten._foreach_add.Scalar,
        aten._foreach_div.Scalar,
        aten._foreach_mul.Scalar,
        aten._foreach_sub.Scalar,
    ])
def _prop__foreach_binop_scalar(op_schema: OpSchema) -> OutputSharding:
    self, scalar = op_schema.args_schema
    assert isinstance(self, list) and all(isinstance(s, DTensorSpec) for s in self)
    assert not isinstance(scalar, list)
    return OutputSharding(output_spec=self)


@register_prop_rule(  # pyre-ignore
    [
        aten._foreach_add.List,
        aten._foreach_div.List,
        aten._foreach_mul.List,
    ])
def _prop__foreach_binop_list(op_schema: OpSchema) -> OutputSharding:
    self, other = op_schema.args_schema[:2]
    scalar = None if len(op_schema.args_schema) < 3 else op_schema.args_schema[2]
    assert isinstance(self, list) and all(
        isinstance(s, DTensorSpec) for s in self), f"Expect a List[DTensorSpec] but got {self}"
    assert isinstance(other, list) and all(
        isinstance(o, DTensorSpec) for o in other), f"Expect a List[DTensorSpec] but got {other}"
    assert len(self) == len(other), ("Two tensor lists must match in length, "
                                     f"but got {len(self)} and {len(other)}")

    # if any(s != o for s, o in zip(self, other)):
    #     # If DTensorSpec for the two operand do not match, suggest using
    #     # self's DTensorSpec. This will trigger allreduce if other is partial
    #     # and self is replicated.
    #     return OutputSharding(
    #         output_spec=None,
    #         schema_suggestions=[
    #             OpSchema(
    #                 func_schema=op_schema.func_schema,
    #                 args_schema=(self, self, scalar) if scalar else (self, self),
    #                 kwargs_schema=op_schema.kwargs_schema,
    #                 is_inplace=op_schema.is_inplace,
    #                 is_out_variant=op_schema.is_out_variant,
    #             )
    #         ],
    #     )
    # else:
    return OutputSharding(output_spec=self)


@register_prop_rule(  # pyre-ignore
    [
        aten._foreach_addcdiv.Scalar,
        aten._foreach_addcmul.Scalar,
    ])
def _prop__foreach_addcop_scalar(op_schema: OpSchema):
    self, tensor1, tensor2 = op_schema.args_schema[:3]
    scalar = None if len(op_schema.args_schema) < 4 else op_schema.args_schema[3]
    assert isinstance(self, list) and all(isinstance(s, DTensorSpec) for s in self)
    assert isinstance(tensor1, list) and all(isinstance(s, DTensorSpec) for s in self)
    assert isinstance(tensor2, list) and all(isinstance(s, DTensorSpec) for s in self)
    if any(s != t1 or s != t2 for s, t1, t2 in zip(self, tensor1, tensor2)):
        # If DTensorSpec for the two operand do not match, suggest using
        # self's DTensorSpec. This will trigger allreduce if other is partial
        # and self is replicated.
        return OutputSharding(
            output_spec=None,
            schema_suggestions=[
                OpSchema(
                    func_schema=op_schema.func_schema,
                    args_schema=(self, self, self, scalar) if scalar else (self, self, self),
                    kwargs_schema=op_schema.kwargs_schema,
                    is_inplace=op_schema.is_inplace,
                    is_out_variant=op_schema.is_out_variant,
                )
            ],
        )
    else:
        return OutputSharding(output_spec=self)


@register_prop_rule(  # pyre-ignore
    [
        aten._foreach_neg.default,
        aten._foreach_reciprocal.default,
        aten._foreach_sqrt.default,
    ])
def _prop__foreach_unaop(op_schema: OpSchema) -> OutputSharding:
    self = op_schema.args_schema[0]
    assert isinstance(self, list) and all(isinstance(s, DTensorSpec) for s in self)
    # FIXME(@mrshenli): for sqrt, this is only mathematically correct for
    # Replicate and Shard tensor.
    return OutputSharding(output_spec=self)


@register_prop_rule(aten.upsample_nearest2d.default)
def _prop_upsample_nearest2d(op_schema: OpSchema) -> OutputSharding:
    input, output_size = op_schema.args_schema[0:2]
    batch, channel = input.shape[:2]
    full_ouput_size = (batch, channel, *output_size)
    return OutputSharding(output_spec=DTensorSpec(
        mesh=input.mesh, placements=input.placements, shape=torch.Size(full_ouput_size)))


@register_prop_rule(aten.upsample_nearest2d_backward.default)
def _prop_upsample_nearest2d_backward(op_schema: OpSchema) -> OutputSharding:
    grad_output, output_size, input_size = op_schema.args_schema[0:3]
    return OutputSharding(output_spec=DTensorSpec(
        mesh=grad_output.mesh, placements=grad_output.placements, shape=torch.Size(input_size)))


@register_prop_rule(aten.tril.default)
def _prop_tril(op_schema: OpSchema) -> OutputSharding:
    self = op_schema.args_schema[0]

    assert isinstance(self, DTensorSpec)
    assert all(isinstance(p, Replicate) for p in self.placements)

    return OutputSharding(output_spec=self)


# (TODO) merge with _tensor/ops/math_ops.py aten._softmax.default
@register_prop_rule(aten._log_softmax.default)
def _prop_log_softmax_rule(op_schema: OpSchema) -> OutputSharding:
    input_spec, softmax_dim, _ = op_schema.args_schema
    input_spec = cast(DTensorSpec, input_spec)
    softmax_dim = cast(int, softmax_dim)
    dim_map = input_spec.dim_map
    if softmax_dim < len(dim_map) and dim_map[softmax_dim] >= 0:
        raise RuntimeError("Cannot run softmax on sharding dimension!")
    return OutputSharding(input_spec)


# (TODO) merge with _tensor/ops/math_ops.py aten._softmax_backward_data.default
@register_prop_rule(aten._log_softmax_backward_data.default)
def softmax_bwd_rule(op_schema: OpSchema) -> OutputSharding:
    grad_out_spec, out_spec, softmax_dim, _ = op_schema.args_schema
    grad_out_spec = cast(DTensorSpec, grad_out_spec)
    out_spec = cast(DTensorSpec, out_spec)
    softmax_dim = cast(int, softmax_dim)
    grad_out_dim_map = grad_out_spec.dim_map
    out_dim_map = out_spec.dim_map
    if softmax_dim < len(grad_out_dim_map) and (grad_out_dim_map[softmax_dim] >= 0
                                                or out_dim_map[softmax_dim] >= 0):
        raise RuntimeError("Cannot run _softmax_backward_data on sharding dimension!")
    return pointwise_rule(op_schema)


@register_prop_rule(aten.nll_loss_forward.default)  # pyre-ignore
def _prop_nll_loss_forward(op_schema: OpSchema) -> OutputSharding:
    self, target = op_schema.args_schema[:2]
    assert isinstance(self, DTensorSpec)
    assert isinstance(target, DTensorSpec)

    out_placement = [Replicate()] * self.mesh.ndim

    for idx, s in enumerate(self.placements):
        if isinstance(s, Shard) and s.dim == 0:
            out_placement[idx] = _Partial()
        else:
            out_placement[idx] = self.placements[idx]

    return OutputSharding(output_spec=(
        DTensorSpec(mesh=self.mesh, placements=out_placement, shape=torch.Size([])),
        DTensorSpec(
            mesh=self.mesh, placements=[Replicate()] * self.mesh.ndim, shape=torch.Size([])),
    ))


@register_prop_rule(aten.nll_loss_backward.default)  # pyre-ignore
def _prop_nll_loss_backward(op_schema: OpSchema) -> OutputSharding:
    grad_output, self = op_schema.args_schema[:2]
    assert isinstance(grad_output, DTensorSpec)
    assert isinstance(self, DTensorSpec)
    return OutputSharding(output_spec=self)


@register_prop_rule(aten.avg_pool3d.default)
def _prop_avg_pool3d(op_schema: OpSchema) -> OutputSharding:
    input, kernel_size, stride = op_schema.args_schema[:3]

    padding = [0, 0, 0]
    if len(op_schema.args_schema) > 3:
        padding = op_schema.args_schema[3]

    output_shape = list(input.shape)

    for inner_dim in range(3):
        output_shape[2 + inner_dim] = int(
            math.floor(input.shape[2 + inner_dim] + 2 * padding[inner_dim] -
                       (kernel_size[inner_dim] - 1) - 1) / stride[inner_dim] + 1)

    return OutputSharding(output_spec=DTensorSpec(
        mesh=input.mesh, placements=input.placements, shape=torch.Size(output_shape)))


# from torch/distributed/_tensor/ops/random_ops.py on pytorch main branch
# copy it because it is not ready in released pytorch
@register_prop_rule(aten.native_dropout.default)
def dropout_rule(op_schema: OpSchema) -> OutputSharding:
    self_spec = cast(DTensorSpec, op_schema.args_schema[0])

    # NOTE: dropout on a partial tensor should be similar to the case of a replicate tensor
    partial = False
    for placement in self_spec.placements:
        if isinstance(placement, _Partial):
            partial = True
            break

    if partial:
        return OutputSharding(
            None,
            failed_reason="aten.native_dropout.default with _Partial is not supported yet!",
        )
    else:
        return OutputSharding(output_spec=(self_spec, self_spec))


@register_prop_rule([aten.max.dim, aten.min.dim])
def reduce_dim_rule(op_schema: OpSchema) -> OutputSharding:
    self, dim = op_schema.args_schema[:2]
    keepdim = False
    if len(op_schema.args_schema) == 3:
        keepdim = op_schema.args_schema[2]

    self_shape, self_placements = list(self.shape), copy.deepcopy(self.placements)
    if keepdim:
        self_shape[dim] = 1
    else:
        del self_shape[dim]

    out_spec = DTensorSpec(mesh=self.mesh,
                           placements=self_placements,
                           shape=self_shape,
                           ndim=self.ndim - 1)

    return OutputSharding(output_spec=(out_spec, out_spec))


@register_prop_rule([
    aten.index_put.default,
    aten.index_put_.default,
])
def reduce_dim_rule(op_schema: OpSchema) -> OutputSharding:
    self = op_schema.args_schema[0]
    return OutputSharding(output_spec=self)


@register_prop_rule(aten._scaled_dot_product_efficient_attention.default)
def scaled_dot_product_efficient_attention_rule(op_schema: OpSchema) -> OutputSharding:
    query, key, value = op_schema.args_schema[0:3]

    out_placements = query.placements
    dim_size = len(value.shape)
    for idx, s in enumerate(value.placements):
        if isinstance(s, Shard) and s.dim == dim_size - 1:
            out_placements[idx] = Shard(s.dim)

    out_spec = DTensorSpec(mesh=query.mesh,
                           placements=out_placements,
                           shape=torch.Size(list(query.shape[:-1]) + list(value.shape[-1:])),
                           ndim=query.ndim)

    state_spec = DTensorSpec(mesh=query.mesh,
                             placements=[Replicate()] * query.mesh.ndim,
                             shape=torch.Size(list(query.shape[:-2]) + [0]),
                             ndim=query.ndim - 1)

    return OutputSharding(output_spec=(out_spec, state_spec))


@register_prop_rule(aten.masked_fill_.Scalar)
def masked_fill_rule(op_schema: OpSchema) -> OutputSharding:
    tensor = op_schema.args_schema[0]
    return OutputSharding(output_spec=tensor)


# from torch/distributed/_tensor/ops/embedding_ops.py on pytorch main branch
# copy it because it is not ready in released pytorch
@register_prop_rule(aten.embedding_dense_backward.default)
def embedding_dense_backward_rules(op_schema: OpSchema) -> OutputSharding:
    grad_output, indices, num_weights = op_schema.args_schema[:3]
    assert isinstance(grad_output, DTensorSpec)
    assert isinstance(indices, DTensorSpec)

    output_shape = (num_weights, ) + grad_output.shape[len(indices.shape):]
    output_placements = []

    for idx in range(len(grad_output.placements)):
        if grad_output.placements[idx] == indices.placements[idx]:
            if isinstance(grad_output.placements[idx], Replicate):
                output_placements.append(Replicate())
            else:
                output_placements.append(_Partial())
        elif grad_output.placements == [_Partial()] and indices.placements == [Replicate()]:
            # The embedding table is replicated and the indices is also replicated
            # (local is a more precise term). This is postional embedding. In this
            # case, gradients for the embmedding table should be Partial.
            output_placements.append(_Partial())
        else:
            raise NotImplementedError("Unsupported embedding dense backward schema:\n"
                                      f"grad_output - {grad_output}\n"
                                      f"indices - {indices}")

    return OutputSharding(output_spec=DTensorSpec(
        mesh=indices.mesh, shape=output_shape, placements=output_placements))


# from torch/distributed/_spmd/experimental_ops.py on pytorch main branch
# copy it because it is not ready in released pytorch
@register_prop_rule(aten.stack.default)
def _prop_stack(op_schema: OpSchema) -> OutputSharding:
    tensors = op_schema.args_schema[0]
    dim = 0 if len(op_schema.args_schema) == 1 else cast(int, op_schema.args_schema[1])
    assert (isinstance(tensors, list) and len(tensors) > 0), "expect at least one tensor to stack"
    assert all(isinstance(t, DTensorSpec)
               for t in tensors), f"expect a list of DTensorSpecs, but got {tensors}"
    assert all(t.shape == tensors[0].shape
               for t in tensors), f"expect all tensors to have the same shape, but got {tensors}."
    # TODO: provide schema_suggestions when placements do not match
    assert all(
        t.placements == tensors[0].placements
        for t in tensors), f"expect all tensors to have the same placements, but got {tensors}."
    assert all(
        not p.is_shard(dim)
        for p in tensors[0].placements), "DTensor does not support stack on sharded dimension."

    concat_placements = copy.deepcopy(tensors[0].placements)

    for idx, s in enumerate(concat_placements):
        if isinstance(s, Shard) and s.dim >= dim:
            concat_placements[idx] = Shard(s.dim + 1)

    output_shape = list(tensors[0].shape)
    output_shape.insert(dim, len(tensors))

    return OutputSharding(output_spec=DTensorSpec(
        mesh=tensors[0].mesh, shape=output_shape, placements=tensors[0].placements))



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import re
from typing import Any

import jax
from jax._src import source_info_util
from jax._src.util import safe_map
from jax._src.core import ShapedArray

from easydist.metashard import MetaOp, view_propagation
from .device_mesh import device_mesh_world_size


def generate_str(invals, bind_params):
    return str([(val.shape, val.dtype.name) if isinstance(val, jax.Array) else val
                for val in invals]) + str(bind_params)


def inject_view_propagation(sharding_info):

    view_op_list = ["reshape"]

    for op_name in sharding_info:
        if op_name in view_op_list:
            for args in sharding_info[op_name]:
                # example args for view:
                # "[((4, 256, 768), dtype('float32'))]{'new_sizes': (1, 4, 256, 768), 'dimensions': None}"
                re_matches = re.findall(r'\((\d+,|\d+(?:, \d+)*)\)', args)
                input_shape = [int(num) for num in re_matches[0].strip(',').split(',')]
                output_shape = [int(num) for num in re_matches[1].strip(',').split(',')]
                sharding_info[op_name][args] = view_propagation(input_shape, 
                                                                output_shape, 
                                                                world_size=device_mesh_world_size())

    return sharding_info


class EDJaxShardingAnn:

    def __init__(self, jaxpr, use_cache=True, seed=42) -> None:
        self.jaxpr = jaxpr
        self.use_cache = use_cache
        self.shape_info = {}
        self.sharding_info = {}

        self.key = jax.random.PRNGKey(seed)

    def run(self, consts, *args) -> Any:
        # Mapping from variable -> value
        env = {}

        def read(var):
            # Literals are values baked into the Jaxpr
            if type(var) is jax.core.Literal:
                return var.val
            if isinstance(env[var], ShapedArray):
                self.key, subkey = jax.random.split(self.key)
                if env[var].dtype.name in ["float64", "float32", "float16"]:
                    return jax.random.normal(subkey, shape=env[var].shape, dtype=env[var].dtype)
                elif env[var].dtype.name in ["int32", "unint32", "int64", "uint64", "uint8"]:
                    return jax.random.randint(subkey,
                                              shape=env[var].shape,
                                              dtype=env[var].dtype,
                                              minval=1,
                                              maxval=8)
                elif env[var].dtype.name in ["bool"]:
                    return jax.random.normal(subkey, shape=env[var].shape) > 1.
                else:
                    return jax.numpy.empty(shape=env[var].shape, dtype=env[var].dtype)
            return env[var]

        def write(var, val):
            if isinstance(val, int):
                env[var] = ShapedArray(shape=set(), dtype=jax.numpy.int32)
            elif isinstance(val, jax.Array) and not jax.core.is_opaque_dtype(val.dtype):
                env[var] = ShapedArray(shape=val.shape, dtype=val.dtype)
            else:
                env[var] = val

        def record_shape(var):
            if isinstance(env[var], jax.Array) or isinstance(env[var], ShapedArray):
                self.shape_info[var.__str__()] = {
                    'shape': env[var].shape,
                    'dtype': env[var].dtype.name
                }

        flat_args, _ = jax.tree_util.tree_flatten(args)

        # Bind args and consts to environment
        safe_map(write, self.jaxpr.invars, flat_args)
        safe_map(write, self.jaxpr.constvars, consts)

        safe_map(record_shape, self.jaxpr.invars)
        safe_map(record_shape, self.jaxpr.constvars)

        # add sharding_info for invars+constvars
        for invar in self.jaxpr.invars + self.jaxpr.constvars:
            invals = env[invar]
            onelevel_key = "placeholder"
            twolevel_key = str([(invals.shape, invals.dtype.name)])
            if onelevel_key not in self.sharding_info:
                self.sharding_info[onelevel_key] = {}
            self.sharding_info[onelevel_key][twolevel_key] = view_propagation(invals.shape, 
                                                                              invals.shape, 
                                                                              world_size=device_mesh_world_size())

        # Loop through equations and evaluate primitives using `bind`
        for eqn in self.jaxpr.eqns:
            subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
            onelevel_key = eqn.primitive.__str__()
            name_stack = source_info_util.current_name_stack() + eqn.source_info.name_stack

            with source_info_util.user_context(eqn.source_info.traceback, name_stack=name_stack):

                # Read inputs to equation from environment
                invals = safe_map(read, eqn.invars)

                if eqn.primitive.__str__() == "custom_jvp_call":
                    pjit_eqn = subfuns[0].f.args[0].eqns[0]
                    onelevel_key += "[" + pjit_eqn.params['name'] + "]"
                elif eqn.primitive.__str__() in ["xla_call", "pjit"]:
                    onelevel_key += "[" + eqn.params['name'] + "]"

                # `bind` is how a primitive is called
                meta_op_ = MetaOp(func=eqn.primitive.bind,
                                  input_args=(subfuns, invals, bind_params),
                                  name=onelevel_key)
                outvals = meta_op_.exec()

            if onelevel_key not in self.sharding_info:
                self.sharding_info[onelevel_key] = {}

            twolevel_key = generate_str(invals, bind_params)

            if not self.use_cache or twolevel_key not in self.sharding_info[onelevel_key]:
                prompt_annotation = None
                if len(self.sharding_info[onelevel_key]) >= 1:
                    prompt_annotation = list(
                        self.sharding_info[onelevel_key].values())[0]["sharding_ann"]
                sharding_ann, combination_ann = meta_op_.sharding_discovery(
                    prompt_annotation=prompt_annotation)
                self.sharding_info[onelevel_key][twolevel_key] = {
                    "sharding_ann": sharding_ann,
                    "combination_ann": combination_ann
                }

            # outvals = eqn.primitive.bind(*invals, **eqn.params)
            # Primitives may return multiple outputs or not
            if not eqn.primitive.multiple_results:
                outvals = [outvals]
            # Write the results of the primitive into the environment
            safe_map(write, eqn.outvars, outvals)
            safe_map(record_shape, eqn.outvars)

        # Read the final result of the Jaxpr from the environment
        # return safe_map(read, self.jaxpr.outvars)
        return inject_view_propagation(self.sharding_info), self.shape_info



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from functools import partial

import jax

add = jax.numpy.add
equal = jax.numpy.array_equal
zeros_like = jax.numpy.zeros_like
min = jax.numpy.minimum
max = jax.numpy.maximum
allclose = partial(jax.numpy.allclose, rtol=5e-3, atol=5e-03)


def concatenate(tensors, dim=0):
    return jax.numpy.concatenate(tensors, axis=dim)


def chunk(input, chunks, dim=0):
    return jax.numpy.array_split(input, chunks, axis=dim)


def narrow(input, dim, start, length):
    indices = jax.numpy.asarray(range(start, start + length))
    return jax.numpy.take(input, indices, axis=dim)


Tensor = jax.Array

tree_flatten = jax.tree_util.tree_flatten


def tree_unflatten(values, spec):
    return jax.tree_util.tree_unflatten(spec, values)


clone = jax.numpy.copy

from_numpy = jax.numpy.array



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
import random
from pprint import pprint
from typing import List

import easydist.config as mdconfig
import mip
from easydist.metashard.metair import (SPMD, MetaGraph, MetaNode, MetaNodeCluster, MetaVar,
                                       NodeSPMDStrategy, VarSPMDStrategy, VarSPMDStrategyGroup,
                                       ClusterStrategyPool)

logger = logging.getLogger(__name__)


def shuffle_list(*ls):
    l = list(zip(*ls))

    random.shuffle(l)
    return zip(*l)


def get_idx_in_var_list(var: MetaVar, var_list: List[MetaVar]):
    var_list = [v.name if v else None for v in var_list]
    if var.name in var_list:
        return var_list.index(var.name)
    return None


def calculate_resharding_cost(var: MetaVar, strategy_in: VarSPMDStrategy,
                              strategy_out: VarSPMDStrategy, device_mesh):
    var_size = var.get_var_size()

    all_gather = lambda x, ndevices: x * (ndevices - 1) / ndevices
    all_reduce = lambda x, ndevices: 2 * x * (ndevices - 1) / ndevices

    def all_to_all(x, ndevices):
        factor = 1.
        if ndevices > 2:
            factor = mdconfig.all_to_all_punish_factor
        return factor * x * (ndevices - 1) / ndevices / ndevices

    resharding_cost = 0

    s1_in, s2_in = strategy_in[0], strategy_in[1]
    s1_out, s2_out = strategy_out[0], strategy_out[1]

    if device_mesh[0] > 1:
        message_size = var_size
        if s2_in.is_shard():
            message_size /= device_mesh[1]
        if s1_in.is_shard():
            if s1_out.is_shard():
                if s1_in.args != s1_out.args:
                    resharding_cost += all_to_all(message_size, device_mesh[0])
            else:
                resharding_cost += all_gather(message_size, device_mesh[0])
        elif s1_in.is_partial():
            resharding_cost += all_reduce(message_size, device_mesh[0])

    if device_mesh[1] > 1:
        message_size = var_size
        if s1_in.is_shard():
            message_size /= device_mesh[0]
        if s2_in.is_shard():
            if s2_out.is_shard():
                if s2_in.args != s2_out.args:
                    resharding_cost += all_to_all(message_size, device_mesh[1])
            else:
                resharding_cost += all_gather(message_size, device_mesh[1])
        elif s2_in.is_partial():
            resharding_cost += all_reduce(message_size, device_mesh[1])

    return resharding_cost


def calculate_memory_cost(var: MetaVar, strategy_in: VarSPMDStrategy,
                          strategy_out: VarSPMDStrategy, device_mesh):
    var_size = var.get_var_size()

    memory_cost = 0

    # FIXME: only use out_strategy here for shard_size is ok?
    for strategy in [strategy_in, strategy_out]:
        s1, s2 = strategy[0], strategy[1]
        shard_size = 1
        if s1.is_shard():
            shard_size *= device_mesh[0]
        if s2.is_shard():
            shard_size *= device_mesh[1]

        memory_cost += var_size // shard_size

    return memory_cost


def gen_comm_cost_matrix(
        var: MetaVar,
        up_strategy,  # list of VarSPMDStrategy
        down_strategy,  # list of VarSPMDStrategy
        device_mesh):
    comm_matrix = [[0 for _ in range(len(down_strategy))] for _ in range(len(up_strategy))]

    for i in range(len(up_strategy)):
        for j in range(len(down_strategy)):
            var_up_strategy = up_strategy[i]
            var_down_strategy = down_strategy[j]
            comm_matrix[i][j] = calculate_resharding_cost(var, var_up_strategy, var_down_strategy,
                                                          device_mesh)

    return comm_matrix


# Lansong(TODO) generate_comm_matrix is deprecated, use gen_comm_cost_matrix instead
def generate_comm_matrix(
        var: MetaVar,
        up_strategy,  # list of NodeSPMDStrategy
        down_strategy,  # list of NodeSPMDStrategy
        idx_for_up,
        idx_for_down,
        device_mesh):
    comm_matrix = [[0 for _ in range(len(down_strategy))] for _ in range(len(up_strategy))]

    for i in range(len(up_strategy)):
        for j in range(len(down_strategy)):
            var_up_strategy = up_strategy[i].out_strtg_group[idx_for_up]
            var_down_strategy = down_strategy[j].in_strtg_group[idx_for_down]
            comm_matrix[i][j] = calculate_resharding_cost(var, var_up_strategy, var_down_strategy,
                                                          device_mesh)

    return comm_matrix


def gen_mem_cost_matrix(var: MetaVar, up_strategy, down_strategy, device_mesh):
    comm_matrix = [[0 for _ in range(len(down_strategy))] for _ in range(len(up_strategy))]

    for i in range(len(up_strategy)):
        for j in range(len(down_strategy)):
            var_up_strategy = up_strategy[i]
            var_down_strategy = down_strategy[j]
            comm_matrix[i][j] = calculate_memory_cost(var, var_up_strategy, var_down_strategy,
                                                      device_mesh)

    return comm_matrix


# Lansong(TODO) generate_mem_matrix is deprecated, use gen_mem_cost_matrix instead
def generate_mem_matrix(var: MetaVar, up_strategy, down_strategy, idx_for_up, idx_for_down,
                        device_mesh):
    comm_matrix = [[0 for _ in range(len(down_strategy))] for _ in range(len(up_strategy))]

    for i in range(len(up_strategy)):
        for j in range(len(down_strategy)):
            var_up_strategy = up_strategy[i].out_strtg_group[idx_for_up]
            var_down_strategy = down_strategy[j].in_strtg_group[idx_for_down]
            comm_matrix[i][j] = calculate_memory_cost(var, var_up_strategy, var_down_strategy,
                                                      device_mesh)

    return comm_matrix


class ClusterMipInfo:

    def __init__(self, cluster: MetaNodeCluster, cluster_strtg_pool: ClusterStrategyPool,
                 mip_var: list):
        self.cluster = cluster
        self.cluster_strtg_pool = cluster_strtg_pool
        self.mip_var = mip_var

    def get_node_strtg(self, nd_id: int, strtg_idx: int) -> NodeSPMDStrategy:
        return self.cluster_strtg_pool.get_node_strtg(nd_id, strtg_idx)

    def __str__(self) -> str:
        res = str(self.cluster)
        res += "\n" + str(self.cluster_strtg_pool)
        return res

    def __repr__(self) -> str:
        return self.__str__()


class ClusterEdgeMipInfo:

    def __init__(self, edge: MetaVar):
        self.edge = edge
        self.up_node = None
        self.down_nodes = []
        self.idx_for_up = None
        self.idx_for_down = []
        self.comm_matrix = []
        self.mem_matrix = []
        self.mip_var = []

    def __str__(self) -> str:
        res = str(self.edge)
        return res

    def __repr__(self) -> str:
        return self.__str__()


class AutoFlowSolver:

    def __init__(self,
                 device_mesh=None,
                 constraints=None,
                 memory_ratio=0.9,
                 total_memery=None) -> None:
        self.m = mip.Model("autoflow")
        self.nodes = {}
        self.edges = {}
        self.device_mesh = device_mesh
        self.constraints = constraints

        self.memory_ratio = memory_ratio
        self.total_memery = 32 * 1024 * 1024 * 1024 if total_memery is None else total_memery
        self.max_memory_constrain = self.memory_ratio * self.total_memery

        self.clusters = {}
        self.cluster_edges = {}

    def add_coarsen_graph(self, graph: MetaGraph) -> None:
        self.graph = graph

        for cluster in graph.node_clusters:
            self.add_cluster(cluster)

        if mdconfig.log_level <= logging.DEBUG:
            for mip_info in self.clusters.values():
                print(str(mip_info))

    def add_cluster(self, cluster: MetaNodeCluster) -> None:
        cluster_id = cluster.unique_id
        cluster_strtg_pool = cluster.get_strtg_pool()
        cluster_strtg_num = cluster_strtg_pool.get_strtg_num()
        if cluster_strtg_num > 0:
            mip_vars = [self.m.add_var(var_type=mip.BINARY) for _ in range(cluster_strtg_num)]
            mip_info = ClusterMipInfo(cluster, cluster_strtg_pool, mip_vars)
            self.clusters[cluster_id] = mip_info

            # Lansong 2do
            # add edges from user side
            for input_node, input_idx in cluster.args.descs:
                if input_node.op_name == "placeholder" or input_node.op_name == "get_attr":
                    # placeholder or get_attr for a stateful variable has an implicit input
                    if input_node in self.graph.state_io_map:
                        output_var = self.graph.state_io_map[input_node]

                        # output var is the implicit input var of placeholder and get_attr
                        self.add_cluster_edge(output_var, input_idx, down_node=input_node)
                else:
                    var = input_node.invars[input_idx]
                    self.add_cluster_edge(var, input_idx, down_node=input_node)

            assert cluster.output_node
            for out_idx, var in enumerate(cluster.output_node.outvars):
                if var:
                    self.add_cluster_edge(var, out_idx, up_node=cluster.output_node)

    # Lansong(TODO) add_graph is deprecated, use add_coarsen_graph instead
    def add_graph(self, graph: MetaGraph) -> None:
        self.graph = graph

        if self.constraints:
            for cons in self.constraints:
                if cons.name not in [i.name for i in self.graph.input_list]:
                    continue
                cons_node = MetaNode("cons_" + cons.name, [], [cons], None)
                outvars_strategy = VarSPMDStrategyGroup(VarSPMDStrategy(self.constraints[cons]))
                self.nodes[cons_node.unique_key()] = {
                    "node": cons_node,
                    "strategy": NodeSPMDStrategy(None, outvars_strategy),
                    "mip_var": [self.m.add_var(var_type=mip.BINARY)]
                }
                self.add_edge(cons, up_node=cons_node)

        self.liveness = graph.liveness()
        if mdconfig.liveness_only_input:
            self.liveness = self.liveness[:1]

        for op in graph.op_list:
            self.add_node(op)

    # Lansong(TODO) add_node is deprecated, use add_cluster instead
    def add_node(self, node: MetaNode) -> None:
        unique_key_ = node.unique_key()

        strategies = node.get_strtg_pool().strategies
        if len(strategies) > 0:
            self.nodes[unique_key_] = {
                "node": node,
                "strategy": strategies,
                "mip_var": [self.m.add_var(var_type=mip.BINARY) for _ in range(len(strategies))]
            }

            for var in node.invars:
                self.add_edge(var, down_node=node)

            for var in node.outvars:
                if var:
                    self.add_edge(var, up_node=node)

    def add_cluster_edge(self, edge: MetaVar, io_idx: int, up_node=None, down_node=None) -> None:
        unique_key_ = edge.name

        if unique_key_ not in self.cluster_edges:
            edge_mip_info = ClusterEdgeMipInfo(edge=edge)
            self.cluster_edges[unique_key_] = edge_mip_info

        if up_node is not None:
            self.cluster_edges[unique_key_].up_node = up_node
            self.cluster_edges[unique_key_].idx_for_up = io_idx

        if down_node is not None:
            self.cluster_edges[unique_key_].down_nodes.append(down_node)
            self.cluster_edges[unique_key_].idx_for_down.append(io_idx)


        def _add_cluster_edge(up_node, idx_for_up, down_node, idx_for_down):
            up_cluster_strtg_pool = self.clusters[up_node.cluster_id].cluster_strtg_pool
            up_out_strategy_list = up_cluster_strtg_pool.get_outvar_strtg_list(
                up_node.unique_id, idx_for_up)

            down_cluster_strtg_pool = self.clusters[down_node.cluster_id].cluster_strtg_pool
            down_in_strategy_list = down_cluster_strtg_pool.get_invar_strtg_list(
                down_node.unique_id, idx_for_down)

            self.cluster_edges[unique_key_].mip_var.append([[
                self.m.add_var(var_type=mip.BINARY) for _ in range(len(down_in_strategy_list))
            ] for _ in range(len(up_out_strategy_list))])

            # calculate ``comm_matrix`` for this edge
            idx_for_up = self.cluster_edges[unique_key_].idx_for_up
            idx_for_down = self.cluster_edges[unique_key_].idx_for_down[-1]

            self.cluster_edges[unique_key_].comm_matrix.append(
                gen_comm_cost_matrix(
                    self.cluster_edges[unique_key_].edge,
                    up_out_strategy_list,
                    down_in_strategy_list,
                    self.device_mesh,
                ))

            self.cluster_edges[unique_key_].mem_matrix.append(
                gen_mem_cost_matrix(
                    self.cluster_edges[unique_key_].edge,
                    up_out_strategy_list,
                    down_in_strategy_list,
                    self.device_mesh,
                ))

        # if adding an up_node forms an edge
        if up_node and self.cluster_edges[unique_key_].down_nodes:
            for down_node, idx_for_down in zip(self.cluster_edges[unique_key_].down_nodes, 
                                               self.cluster_edges[unique_key_].idx_for_down):
                _add_cluster_edge(up_node, io_idx, down_node, idx_for_down)
        # if adding a down_node forms an edge
        elif self.cluster_edges[unique_key_].up_node and down_node:
            _add_cluster_edge(self.cluster_edges[unique_key_].up_node,
                              self.cluster_edges[unique_key_].idx_for_up,
                              down_node, io_idx)


    # Lansong(TODO) add_edge is deprecated, use add_cluster_edge instead
    def add_edge(self, edge: MetaVar, up_node=None, down_node=None) -> None:
        unique_key_ = edge.name

        if unique_key_ not in self.edges:
            self.edges[unique_key_] = {
                "edge": edge,
                "up_node": None,
                "down_node": [],
                "idx_for_up": None,
                "idx_for_down": [],
                "comm_matrix": [],
                "mem_matrix": [],
                "mip_var": [],
            }
        if up_node is not None:
            self.edges[unique_key_]["up_node"] = up_node.unique_key()
            self.edges[unique_key_]["idx_for_up"] = get_idx_in_var_list(edge, up_node.outvars)

        if down_node is not None:
            self.edges[unique_key_]["down_node"].append(down_node.unique_key())
            self.edges[unique_key_]["idx_for_down"].append(
                get_idx_in_var_list(edge, down_node.invars))

            if self.edges[unique_key_]["up_node"] is not None:

                up_node_key = self.edges[unique_key_]["up_node"]
                up_strategy = self.nodes[up_node_key]["strategy"]

                down_node_key = self.edges[unique_key_]["down_node"][-1]
                down_strategy = self.nodes[down_node_key]["strategy"]

                self.edges[unique_key_]["mip_var"].append(
                    [[self.m.add_var(var_type=mip.BINARY) for _ in range(len(down_strategy))]
                     for _ in range(len(up_strategy))])

                # calculate ``comm_matrix`` for this edge
                idx_for_up = self.edges[unique_key_]["idx_for_up"]
                idx_for_down = self.edges[unique_key_]["idx_for_down"][-1]

                self.edges[unique_key_]["comm_matrix"].append(
                    generate_comm_matrix(
                        self.edges[unique_key_]["edge"],
                        up_strategy,  # list of NodeSPMDStrategy
                        down_strategy,  # list of NodeSPMDStrategy
                        idx_for_up,
                        idx_for_down,
                        self.device_mesh,
                    ))

                self.edges[unique_key_]["mem_matrix"].append(
                    generate_mem_matrix(
                        self.edges[unique_key_]["edge"],
                        up_strategy,  # list of NodeSPMDStrategy
                        down_strategy,  # list of NodeSPMDStrategy
                        idx_for_up,
                        idx_for_down,
                        self.device_mesh,
                    ))

    # Lansong(TODO) ilp_optimize is deprecated, use ilp_solve instead
    def ilp_optimize(self, count_invars=False):
        comm_cost, mem_cost = 0, 0
        for edge in self.edges.values():
            for idx in range(len(edge["mip_var"])):
                mip_var = edge["mip_var"][idx]
                comm_matrix = edge["comm_matrix"][idx]
                mem_matrix = edge["mem_matrix"][idx]
                shape_1 = len(mip_var)
                shape_2 = len(mip_var[0])
                comm_cost = comm_cost + mip.xsum(mip_var[i][j] * comm_matrix[i][j]
                                                 for i in range(shape_1) for j in range(shape_2))
                mem_cost = mem_cost + mip.xsum(mip_var[i][j] * mem_matrix[i][j]
                                               for i in range(shape_1) for j in range(shape_2))

            def _mem_cost(var_size, down_strategy, idx_for_down):
                memory_cost_list = []
                for i in range(len(down_strategy)):
                    strategy = down_strategy[i].in_strtg_group[idx_for_down]

                    # FIXME: only use out_strategy here for shard_size is ok?
                    s1, s2 = strategy[0], strategy[1]
                    shard_size = 1
                    if s1.state == SPMD.SHARD:
                        shard_size *= self.device_mesh[0]
                    if s2.state == SPMD.SHARD:
                        shard_size *= self.device_mesh[1]

                    memory_cost_list.append(var_size // shard_size)

                return memory_cost_list

            if count_invars and edge["up_node"] is None:
                var_size = edge["edge"].get_var_size()
                for down_node_key, idx_for_down in zip(edge["down_node"], edge["idx_for_down"]):
                    down_strategy = self.nodes[down_node_key]["strategy"]
                    __mem_cost = _mem_cost(var_size, down_strategy, idx_for_down)
                    down_node_mip_var = self.nodes[down_node_key]["mip_var"]
                    mem_cost = mem_cost + mip.xsum(down_node_mip_var[i] * __mem_cost[i]
                                                   for i in range(len(down_node_mip_var)))

        for edge in self.edges.values():
            for idx in range(len(edge["mip_var"])):
                mip_var = edge["mip_var"][idx]
                shape_1 = len(mip_var)
                shape_2 = len(mip_var[0])
                self.m += mip.xsum(mip_var[i][j] for i in range(shape_1)
                                   for j in range(shape_2)) == 1

                up_node_key = edge["up_node"]
                up_node_mip_var = self.nodes[up_node_key]["mip_var"]

                down_node_key = edge["down_node"][idx]
                down_node_mip_var = self.nodes[down_node_key]["mip_var"]

                for i in range(shape_1):
                    for j in range(shape_2):
                        self.m += mip_var[i][j] <= up_node_mip_var[i]
                        self.m += mip_var[i][j] <= down_node_mip_var[j]
                        self.m += mip_var[i][j] >= up_node_mip_var[i] + down_node_mip_var[j] - 1

        idx_record = {}
        for idx_ in range(len(self.liveness)):
            mem_live = self.liveness[idx_]
            op_ = self.graph.op_list[idx_]

            mip_var_list = []
            mem_matrix_list = []
            for tensor_name in mem_live:
                if tensor_name not in self.edges:
                    continue
                if tensor_name not in idx_record:
                    idx_record[tensor_name] = 0
                if len(self.edges[tensor_name]["mip_var"]) > 0:
                    edge_idx = min(
                        len(self.edges[tensor_name]["mip_var"]) - 1, idx_record[tensor_name])
                    mip_var_list.append(self.edges[tensor_name]["mip_var"][edge_idx])
                    mem_matrix_list.append(self.edges[tensor_name]["mem_matrix"][edge_idx])
                if self.edges[tensor_name]["up_node"] is None:
                    down_node_key, idx_for_down = self.edges[tensor_name]["down_node"][
                        0], self.edges[tensor_name]["idx_for_down"][0]
                    var_size = self.edges[tensor_name]["edge"].get_var_size()
                    down_strategy = self.nodes[down_node_key]["strategy"]
                    __mem_cost = _mem_cost(var_size, down_strategy, idx_for_down)
                    mip_var_list.append([self.nodes[down_node_key]["mip_var"]])
                    mem_matrix_list.append([__mem_cost])

            for var in op_.invars:
                if var.name not in self.edges:
                    continue
                if var.name not in idx_record:
                    idx_record[var.name] = 0
                idx_record[var.name] += 1

            if len(mip_var_list) >= 1:
                need_sum = []
                for mip_var, mem_matrix in zip(mip_var_list, mem_matrix_list):
                    shape_1 = len(mip_var)
                    shape_2 = len(mip_var[0])
                    for i in range(shape_1):
                        for j in range(shape_2):
                            need_sum.append(mip_var[i][j] * mem_matrix[i][j])
                self.m += mip.xsum(i for i in need_sum) <= self.max_memory_constrain

        for node in self.nodes.values():
            mip_var = node["mip_var"]
            shape_1 = len(mip_var)
            self.m += mip.xsum(mip_var[i] for i in range(shape_1)) == 1

        self.m.objective = mip.minimize(comm_cost + 0.00000001 * mem_cost)

        self.m.verbose = 0
        status = self.m.optimize(max_seconds_same_incumbent=mdconfig.max_seconds_same_incumbent)
        logger.info(f'[AutoFlowSolver.status]:\t {status}')
        logger.info(f'[AutoFlowSolver.solution_cost]:\t {self.m.objective_value}')

        return self.get_optimal_stratey()

    # Lansong(TODO) get_optimal_stratey is deprecated
    def get_optimal_stratey(self):
        optimal_stratey = {}
        for unique_key_ in self.nodes:
            node = self.nodes[unique_key_]['node']
            opt_ = None
            strategy_list = self.nodes[unique_key_]['strategy']
            mip_var = self.nodes[unique_key_]['mip_var']
            assert len(strategy_list) == len(mip_var)
            for s_, mip_var_s in zip(strategy_list, mip_var):
                if mip_var_s.x == 1:
                    opt_ = s_
            optimal_stratey[unique_key_] = {'node': node, 'strategy': opt_}

        if mdconfig.log_level <= logging.DEBUG:
            print("optimal_stratey:")
            pprint(optimal_stratey)

        return optimal_stratey

    def ilp_solve(self, count_invars=False):
        comm_cost, mem_cost = 0, 0
        for edge_mip_info in self.cluster_edges.values():
            for idx in range(len(edge_mip_info.mip_var)):
                mip_var = edge_mip_info.mip_var[idx]
                comm_matrix = edge_mip_info.comm_matrix[idx]
                mem_matrix = edge_mip_info.mem_matrix[idx]
                shape_1 = len(mip_var)
                shape_2 = len(mip_var[0])
                comm_cost = comm_cost + mip.xsum(mip_var[i][j] * comm_matrix[i][j]
                                                 for i in range(shape_1) for j in range(shape_2))
                mem_cost = mem_cost + mip.xsum(mip_var[i][j] * mem_matrix[i][j]
                                               for i in range(shape_1) for j in range(shape_2))

            def _mem_cost(var_size, down_strategy: 'list[VarSPMDStrategy]'):
                memory_cost_list = []
                for i in range(len(down_strategy)):
                    strategy = down_strategy[i]

                    # FIXME: only use out_strategy here for shard_size is ok?
                    s1, s2 = strategy[0], strategy[1]
                    shard_size = 1
                    if s1.state == SPMD.SHARD:
                        shard_size *= self.device_mesh[0]
                    if s2.state == SPMD.SHARD:
                        shard_size *= self.device_mesh[1]

                    memory_cost_list.append(var_size // shard_size)

                return memory_cost_list

            if count_invars and edge_mip_info.up_node is None:
                var_size = edge_mip_info.edge.get_var_size()
                for down_node, idx_for_down in zip(edge_mip_info.down_nodes,
                                                   edge_mip_info.idx_for_down):
                    down_cluster_mip_info = self.clusters[down_node.cluster_id]
                    down_cluster_strtg_pool = down_cluster_mip_info.cluster_mip_info
                    down_strategy = down_cluster_strtg_pool.get_invar_strtg_list(
                        down_node.unique_id, idx_for_down)
                    __mem_cost = _mem_cost(var_size, down_strategy)
                    down_cluster_mip_var = self.clusters[down_node.cluster_id].mip_var
                    mem_cost = mem_cost + mip.xsum(down_cluster_mip_var[i] * __mem_cost[i]
                                                   for i in range(len(down_cluster_mip_var)))

        for edge_mip_info in self.cluster_edges.values():
            for idx in range(len(edge_mip_info.mip_var)):
                mip_var = edge_mip_info.mip_var[idx]
                shape_1 = len(mip_var)
                shape_2 = len(mip_var[0])
                self.m += mip.xsum(mip_var[i][j] for i in range(shape_1)
                                   for j in range(shape_2)) == 1

                up_cluster_id = edge_mip_info.up_node.cluster_id
                up_cluster_mip_var = self.clusters[up_cluster_id].mip_var

                down_node = edge_mip_info.down_nodes[idx]
                down_cluster_mip_var = self.clusters[down_node.cluster_id].mip_var

                for i in range(shape_1):
                    for j in range(shape_2):
                        self.m += mip_var[i][j] <= up_cluster_mip_var[i]
                        self.m += mip_var[i][j] <= down_cluster_mip_var[j]
                        self.m += mip_var[i][
                            j] >= up_cluster_mip_var[i] + down_cluster_mip_var[j] - 1


#        Lansong(TODO) support memory limit
#        idx_record = {}
#        for idx_ in range(len(self.liveness)):
#            mem_live = self.liveness[idx_]
#            op_ = self.graph.op_list[idx_]
#
#            mip_var_list = []
#            mem_matrix_list = []
#            for tensor_name in mem_live:
#                if tensor_name not in self.cluster_edges:
#                    continue
#                if tensor_name not in idx_record:
#                    idx_record[tensor_name] = 0
#                if len(self.cluster_edges[tensor_name]["mip_var"]) > 0:
#                    edge_idx = min(
#                        len(self.cluster_edges[tensor_name]["mip_var"]) - 1, idx_record[tensor_name])
#                    mip_var_list.append(self.cluster_edges[tensor_name]["mip_var"][edge_idx])
#                    mem_matrix_list.append(self.cluster_edges[tensor_name]["mem_matrix"][edge_idx])
#                if self.cluster_edges[tensor_name]["up_node"] is None:
#                    down_node_key, idx_for_down = self.cluster_edges[tensor_name]["down_nodes"][
#                        0], self.cluster_edges[tensor_name]["idx_for_down"][0]
#                    var_size = self.cluster_edges[tensor_name]["edge"].get_var_size()
#                    down_strategy = self.clusters[down_node_key]["strategy"]
#                    __mem_cost = _mem_cost(var_size, down_strategy, idx_for_down)
#                    mip_var_list.append([self.clusters[down_node_key]["mip_var"]])
#                    mem_matrix_list.append([__mem_cost])
#
#            for var in op_.invars:
#                if var.name not in self.cluster_edges:
#                    continue
#                if var.name not in idx_record:
#                    idx_record[var.name] = 0
#                idx_record[var.name] += 1
#
#            if len(mip_var_list) >= 1:
#                need_sum = []
#                for mip_var, mem_matrix in zip(mip_var_list, mem_matrix_list):
#                    shape_1 = len(mip_var)
#                    shape_2 = len(mip_var[0])
#                    for i in range(shape_1):
#                        for j in range(shape_2):
#                            need_sum.append(mip_var[i][j] * mem_matrix[i][j])
#                self.m += mip.xsum(i for i in need_sum) <= self.max_memory_constrain

        for cluster_mip_info in self.clusters.values():
            mip_var = cluster_mip_info.mip_var
            shape_1 = len(mip_var)
            self.m += mip.xsum(
                mip_var[i]
                for i in range(shape_1)) == 1  # one and only one cluster strategy is active

        self.m.objective = mip.minimize(comm_cost + 0.00000001 * mem_cost)

        self.m.verbose = 0
        status = self.m.optimize(max_seconds_same_incumbent=mdconfig.max_seconds_same_incumbent)
        logger.info(f'[AutoFlowSolver.status]:\t {status}')
        logger.info(f'[AutoFlowSolver.solution_cost]:\t {self.m.objective_value}')

        return self.get_strategies()

    def get_strategies(self):
        optimal_strategies = {}

        for cluster_mip_info in self.clusters.values():
            cluster = cluster_mip_info.cluster
            opt_strtg_idx = -1
            for strtg_idx, mip_var_s in enumerate(cluster_mip_info.mip_var):
                if mip_var_s.x == 1:
                    opt_strtg_idx = strtg_idx
                    break
            assert opt_strtg_idx >= 0

            for node in cluster.nodes.values():
                nd_strtg = cluster_mip_info.get_node_strtg(node.unique_id, opt_strtg_idx)
                optimal_strategies[node.unique_key()] = {'node': node, 'strategy': nd_strtg}

        if mdconfig.log_level <= logging.DEBUG:
            print("optimal_strategies:")
            pprint(optimal_strategies)

        return optimal_strategies

    def beam_search(self, candidate_num=100):

        def get_new_cost(strategy, node, strategy_idx):
            cost = 0.
            edge_list = [self.edges[invar.name] for invar in node['node'].invars]
            for edge in edge_list:
                up_node_key = edge["up_node"]

                idx = edge["down_node"].index(node['node'].unique_key())
                if up_node_key in strategy:
                    up_node_strategy_idx = strategy[up_node_key]['strategy_idx']
                    mem_cost = edge["mem_matrix"][idx][up_node_strategy_idx][strategy_idx]
                    comm_cost = edge["comm_matrix"][idx][up_node_strategy_idx][strategy_idx]
                    cost += comm_cost

            return cost

        def add_candidate(strategy_candidate, accumulate_cost, node):
            new_strategy_candidate = []
            new_accumulate_cost = []
            key_ = node['node'].unique_key()
            if len(strategy_candidate) == 0:
                for idx in range(len(node['strategy'])):
                    stratey = {key_: {'node': node['node'], 'strategy_idx': idx}}
                    new_strategy_candidate.append(stratey)
                    new_accumulate_cost.append(0.)
            else:
                for idx, strategy in enumerate(strategy_candidate):
                    old_cost = accumulate_cost[idx]
                    for idx in range(len(node['strategy'])):
                        new_cost = get_new_cost(strategy, node, idx)
                        new_strategy = {key: strategy[key] for key in strategy}
                        new_strategy[key_] = {'node': node['node'], 'strategy_idx': idx}
                        new_strategy_candidate.append(new_strategy)
                        new_accumulate_cost.append(old_cost + new_cost)

            return new_strategy_candidate, new_accumulate_cost

        def select_candidate(strategy_candidate, accumulate_cost, candidate_num):
            assert len(strategy_candidate) == len(accumulate_cost)
            if len(accumulate_cost) <= candidate_num:
                return strategy_candidate, accumulate_cost

            accumulate_cost, strategy_candidate = shuffle_list(accumulate_cost, strategy_candidate)
            accumulate_cost, strategy_candidate = zip(
                *sorted(zip(accumulate_cost, strategy_candidate), key=lambda x: x[0]))

            return strategy_candidate[:candidate_num], accumulate_cost[:candidate_num]

        strategy_candidate = []
        accumulate_cost = []
        for unique_key_ in self.nodes:
            node = self.nodes[unique_key_]

            strategy_candidate, accumulate_cost = add_candidate(strategy_candidate,
                                                                accumulate_cost, node)

            strategy_candidate, accumulate_cost = select_candidate(strategy_candidate,
                                                                   accumulate_cost, candidate_num)

            accumulate_cost, strategy_candidate = zip(
                *sorted(zip(accumulate_cost, strategy_candidate), key=lambda x: x[0]))

        strategy = strategy_candidate[0]

        optimal_stratey = {}
        for key in strategy:
            node = strategy[key]['node']
            node_strategy_list = self.nodes[node.unique_key()]['strategy']
            optimal_stratey[key] = {
                "node": node,
                "strategy": node_strategy_list[strategy[key]['strategy_idx']]
            }

        logger.info(f'=========== solution cost:\t {accumulate_cost[0]}')

        return optimal_stratey



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
import os
import pickle
import time
from functools import partial
from typing import Any, cast
from contextlib import nullcontext

import numpy
import rich
import torch
import torch.utils._pytree as pytree
from torch._subclasses.fake_tensor import FakeTensor, FakeTensorMode
from torch.distributed._tensor import (DeviceMesh, DTensor, Replicate, distribute_tensor)
from torch.fx._pytree import tree_flatten_spec
from torch.fx.experimental.proxy_tensor import make_fx
from torch.nn.utils import stateless

import easydist.config as mdconfig
from easydist.autoflow.solver import AutoFlowSolver
from easydist.metashard import metair
from easydist.torch.bridge import (get_torch_sharding_strategy, to_torch_spmd, torch2meta_graph)
from easydist.torch.experimental.decomp_utils import EASYDIST_DECOMP_TABLE
from easydist.torch.experimental.init_helper import (SetParaInitHelper, init_contiguous_buf,
                                                     materialize_zero)
from easydist.torch.passes import (eliminate_detach, fix_addmm_bias, fix_convoluation_bias,
                                   fix_embedding, fix_meta_device, sharding_transform)
from easydist.torch.device_mesh import get_device_mesh, set_device_mesh
from easydist.torch.mem_anaylize import mem_anaylize
from easydist.torch.sharding_interpreter import EDTorchShardingAnn
from easydist.torch.utils import (_enable_compile, _rematerialize_optimizer, _sharding_ann_env)
from easydist.utils import rgetattr, rsetattr
from easydist.utils.testing import TorchMockDeviceMesh

# for pickle dump opt_strategy
import sys

sys.setrecursionlimit(100000)

logger = logging.getLogger(__name__)


def preprocess_traced_graph(fx_module: torch.fx.GraphModule):
    fx_module = fix_meta_device(fx_module)
    fx_module = fix_embedding(fx_module)
    fx_module = fix_addmm_bias(fx_module)
    # fx_module = fix_convoluation_bias(fx_module)
    fx_module = eliminate_detach(fx_module)

    fx_module.recompile()

    return fx_module


def easydist_shard(fx_module: torch.fx.GraphModule, state_tensor_num, *args, **kwargs):

    # (1) preprocess pass
    fx_module = preprocess_traced_graph(fx_module)

    if mdconfig.log_level <= logging.DEBUG:
        fx_module.print_readable()

    # (2) sharding annotation
    with _sharding_ann_env():
        start_t = time.perf_counter()
        sharding_interpreter = EDTorchShardingAnn(fx_module)
        flatten_args = tree_flatten_spec(list(args) + list(kwargs.values()), fx_module._in_spec)
        sharding_info, shape_info = sharding_interpreter.run(*flatten_args)
        logger.info(f"[EDTorchShardingAnn.time]:\t {time.perf_counter() - start_t} s.")
        if mdconfig.log_level <= logging.DEBUG:
            rich.print("sharding_info:\n", sharding_info)
            rich.print("shape_info:\n", shape_info)

    # sync sharding info for all process
    torch.distributed.broadcast_object_list(sharding_info, src=0, device="cuda")

    # (3) translate fx.GraphModule into MetaGraph
    meta_graph = torch2meta_graph(fx_module, state_tensor_num, sharding_info, shape_info)
    meta_graph.dump()

    if mdconfig.log_level <= logging.DEBUG:
        rich.print(meta_graph)

    # (4) construct AutoFlowSolver and run ILP
    device_mesh = get_device_mesh()
    device_mesh_shape = (device_mesh.size(0), device_mesh.size(1))

    total_memery = torch.cuda.get_device_properties(torch.cuda.current_device()).total_memory
    solver = AutoFlowSolver(device_mesh_shape, total_memery=total_memery)

    if mdconfig.enable_graph_coarsen:
        logger.info(f"enable graph coarsen with level {mdconfig.coarsen_level}.")
        solver.add_coarsen_graph(meta_graph)
    else:
        solver.add_graph(meta_graph)

    start_t = time.perf_counter()
    if mdconfig.enable_graph_coarsen:
        opt_strategy = solver.ilp_solve()
    else:
        opt_strategy = solver.ilp_optimize()
    logger.info(f"[AutoFlowSolver.time]:\t {time.perf_counter() - start_t} s.")

    if mdconfig.log_level <= logging.DEBUG:
        rich.print(opt_strategy)

    sharding_strategies = get_torch_sharding_strategy(fx_module, opt_strategy)

    if mdconfig.log_level <= logging.DEBUG:
        rich.print(sharding_strategies)

    return shape_info, meta_graph, opt_strategy, sharding_strategies


@torch.no_grad()
def sharded_tensor(tensor, strategy, mesh, materialize_fn):
    # if tensor is DTensor, redistribute it
    if isinstance(tensor, DTensor):
        return tensor.redistribute(mesh, strategy)

    # materialize FakeTensor and distribute_tensor
    if isinstance(tensor, torch._subclasses.fake_tensor.FakeTensor):
        tensor = materialize_fn(tensor)
    if tensor.is_meta:
        tensor = materialize_fn(tensor)
    return distribute_tensor(tensor, mesh, strategy)


@torch.no_grad()
def dtensor_to_tensor(leaf):
    if isinstance(leaf, DTensor):
        replicate_leaf = leaf.redistribute(leaf.device_mesh, [Replicate()] * len(leaf.placements))
        return replicate_leaf.to_local()
    return leaf


def _compile(func, tracing_mode, init_helper, input_signature, args, kwargs):

    module, opt = None, None

    for arg in pytree.tree_flatten(list(args) + list(kwargs.values()))[0]:
        if isinstance(arg, torch.nn.Module):
            assert module is None, "Only support single nn.Module in args now"
            module = arg
        if isinstance(arg, torch.optim.Optimizer):
            assert opt is None, "Only support single Optimizer in args now"
            opt = arg

    params, buffers = {}, {}
    if module is not None:
        params = dict(module.named_parameters())
        buffers = dict(module.named_buffers())

        if isinstance(init_helper, SetParaInitHelper):
            init_helper.module = module

    named_states = {}

    if opt is not None:
        # assign grad and warm up optimizer
        mode = nullcontext()
        for name in dict(module.named_parameters()):
            with torch.no_grad():
                rsetattr(module, name + ".grad", torch.zeros_like(rgetattr(module, name).data))
                if isinstance(rgetattr(module, name).data, FakeTensor):
                    mode = FakeTensorMode()
        with mode:
            opt.step()
            opt.zero_grad(True)

        for n, p in params.items():
            if p in opt.state:
                named_states[n] = opt.state[p]  # type: ignore[index]
                # if step in state, reduce one for warmup step.
                if 'step' in named_states[n]:
                    named_states[n]['step'] -= 1

    flat_named_states, named_states_spec = pytree.tree_flatten(named_states)

    # fix for sgd withtout momentum
    if all(state is None for state in flat_named_states):
        named_states = {}
        flat_named_states, named_states_spec = pytree.tree_flatten(named_states)

    state_tensor_num = len(params) + len(buffers) + len(flat_named_states)

    def stateless_func(func, params, buffers, named_states, args, kwargs):
        with stateless._reparametrize_module(
                cast(torch.nn.Module, module), {
                    **params,
                    **buffers
                }, tie_weights=True) if module else nullcontext(), _rematerialize_optimizer(
                    opt, named_states, params) if opt else nullcontext():
            ret = func(*args, **kwargs)

        grads = {k: v.grad for k, v in params.items()}
        return params, buffers, named_states, grads, ret

    with _enable_compile():
        traced_graph = make_fx(partial(stateless_func, func),
                               tracing_mode=tracing_mode,
                               decomposition_table=EASYDIST_DECOMP_TABLE,
                               _allow_non_fake_inputs=False)(params, buffers, named_states, args,
                                                             kwargs)

    traced_graph.graph.eliminate_dead_code()
    traced_graph = preprocess_traced_graph(traced_graph)
    traced_graph.recompile()

    if mdconfig.enable_compile_cache:
        os.makedirs(mdconfig.compile_cache_dir, exist_ok=True)
        compiled_cache_file = os.path.join(mdconfig.compile_cache_dir, f"./{input_signature}.pkl")

    if mdconfig.enable_compile_cache and os.path.exists(compiled_cache_file):
        logger.info(f"load compiled result from {compiled_cache_file}.")
        shape_info, sharding_strategy, opt_strategy, args_strategy = pickle.load(
            open(compiled_cache_file, "rb"))
        # mem_anaylize(traced_graph, shape_info, opt_strategy)
        sharded_graph = sharding_transform(traced_graph, sharding_strategy)
        sharded_graph = fix_embedding(sharded_graph, recover=True)
    else:
        shape_info, meta_graph, opt_strategy, sharding_strategy = easydist_shard(
            traced_graph, state_tensor_num, params, buffers, named_states, args, kwargs)
        # mem_anaylize(traced_graph, shape_info, opt_strategy)
        sharded_graph = sharding_transform(traced_graph, sharding_strategy)
        sharded_graph = fix_embedding(sharded_graph, recover=True)
        args_strategy = meta_graph.get_input_strategy(opt_strategy)
        args_strategy = [[to_torch_spmd(i) for i in var_strategy]
                         for var_strategy in args_strategy]

        if mdconfig.enable_compile_cache and torch.distributed.get_rank() == 0:
            pickle.dump([shape_info, sharding_strategy, opt_strategy, args_strategy],
                        open(compiled_cache_file, "wb"))
            logger.info(f"compiled result saved in {compiled_cache_file}.")

    # do not use mock device after get sharded_graph
    device_mesh = get_device_mesh()
    if isinstance(device_mesh, TorchMockDeviceMesh):
        if device_mesh.debug_only:
            world_size = torch.distributed.get_world_size()
            mesh_shape = numpy.array(range(world_size)).reshape(1, -1).tolist()
        else:
            mesh_shape = device_mesh.shape
        mesh = DeviceMesh("cuda", mesh_shape)
        set_device_mesh(mesh)

    device = mdconfig.easydist_device

    # keep fake params, buffers, named_states
    fake_tensor_mode = FakeTensorMode()

    def wrap_fake(x):
        if isinstance(x, torch.Tensor):
            return fake_tensor_mode.from_tensor(x)
        return x

    fake_params = pytree.tree_map(wrap_fake, params)
    fake_buffers = pytree.tree_map(wrap_fake, buffers)
    fake_named_states = pytree.tree_map(wrap_fake, named_states)

    # pre-shard params, buffers, named_states
    params_strategy = args_strategy[:len(params)]
    buffers_strategy = args_strategy[len(params):len(params) + len(buffers)]

    if mdconfig.use_contiguous_buffer:
        contiguous_buf = init_contiguous_buf(params, params_strategy, device_mesh)

    index = 0
    for idx, param_name in enumerate(params):
        materialize_fn = init_helper.get_materialize_fn()
        materialize_fn = partial(materialize_fn,
                                 param_buf_key=param_name,
                                 materialization_device=device)
        params[param_name] = sharded_tensor(params[param_name],
                                            params_strategy[idx],
                                            get_device_mesh(),
                                            materialize_fn=materialize_fn)

        size = params[param_name]._local_tensor.numel()

        if mdconfig.use_contiguous_buffer:
            contiguous_buf[index:index + size] = params[param_name]._local_tensor.view(-1)
            params[param_name]._local_tensor = contiguous_buf[index:index + size].view(
                params[param_name]._local_tensor.shape)

        index += size

    for idx, buffer_name in enumerate(buffers):
        materialize_fn = init_helper.get_materialize_fn()
        materialize_fn = partial(materialize_fn,
                                 param_buf_key=buffer_name,
                                 materialization_device=device)
        buffers[buffer_name] = sharded_tensor(buffers[buffer_name],
                                              buffers_strategy[idx],
                                              get_device_mesh(),
                                              materialize_fn=materialize_fn)

    # use zero init for optimizer states
    flat_named_states, named_states_spec = pytree.tree_flatten(named_states)
    state_tensor_num = len(params) + len(buffers)
    materialize_fn = partial(materialize_zero, materialization_device=device)
    for i in range(len(flat_named_states)):
        if isinstance(flat_named_states[i], torch.Tensor):
            flat_named_states[i] = sharded_tensor(flat_named_states[i],
                                                  args_strategy[state_tensor_num],
                                                  get_device_mesh(),
                                                  materialize_fn=materialize_fn)
            state_tensor_num += 1

    named_states = pytree.tree_unflatten(flat_named_states, named_states_spec)

    @torch.no_grad()
    def compiled_func_return(graph, *args, **kwargs):
        nonlocal params, buffers, named_states

        # tensor to DTensor
        # (TODO) the redistributed of params, buffers, named_states should in graph
        flatten_args, args_specs = pytree.tree_flatten([args, kwargs])

        device = mdconfig.easydist_device
        materialize_fn = partial(materialize_zero, materialization_device=device)

        args_strategy_idx = state_tensor_num
        for i in range(len(flatten_args)):
            if isinstance(flatten_args[i], torch.Tensor):
                flatten_args[i] = sharded_tensor(flatten_args[i],
                                                 args_strategy[args_strategy_idx],
                                                 get_device_mesh(),
                                                 materialize_fn=materialize_fn)
                args_strategy_idx += 1

        args, kwargs = pytree.tree_unflatten(flatten_args, args_specs)

        # run the sharded_graph
        params, buffers, named_states, grads, sharded_out = graph(params, buffers, named_states,
                                                                  args, kwargs)

        for para_name in params:
            params[para_name].grad = grads[para_name]

        # out from DTensor to Tensor
        local_out = pytree.tree_map(dtensor_to_tensor, sharded_out)

        return local_out

    class EDCompiledFunc:

        def __init__(self, func, graph) -> None:
            self.func = func
            self.graph = graph

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            return self.func(self.graph, *args, **kwargs)

        def run_with_graph(self, graph, *args: Any, **kwargs: Any) -> Any:
            return self.func(graph, *args, **kwargs)

        def get_state(self):
            return params, buffers, named_states

        def compile_mono_graph(self, *args: Any, **kwargs: Any):
            with _enable_compile():
                fx_module = make_fx(partial(stateless_func, func),
                                    tracing_mode=tracing_mode,
                                    decomposition_table=EASYDIST_DECOMP_TABLE,
                                    _allow_non_fake_inputs=False)(fake_params, fake_buffers,
                                                                  fake_named_states, args, kwargs)

            fx_module.graph.eliminate_dead_code()
            fx_module.recompile()

            fx_module = preprocess_traced_graph(fx_module)

            sharded_fx_module = sharding_transform(fx_module, sharding_strategy)
            sharded_fx_module = fix_embedding(sharded_fx_module, recover=True)

            if mdconfig.log_level <= logging.DEBUG:
                sharded_fx_module.print_readable()

            return sharded_fx_module

        def parameters(self):
            return params.values()

        def named_parameters(self):
            return params

    # release all cuda memory from module here
    # the param maintain in the local of compiled function.
    if module is not None:
        module.to("meta")

    return EDCompiledFunc(compiled_func_return, sharded_graph)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
import functools
import operator

import numpy
import torch
from torch.distributed._tensor import DeviceMesh

from easydist.metashard import metair
from easydist.utils.testing import TorchMockDeviceMesh

logger = logging.getLogger(__name__)

TORCH_DEVICE_MESH = None


def set_device_mesh(device_mesh):
    global TORCH_DEVICE_MESH
    TORCH_DEVICE_MESH = device_mesh

    if device_mesh.size(0) == 1:
        metair.DEVICE_MESH_1D = 0
    elif device_mesh.size(1) == 1:
        metair.DEVICE_MESH_1D = 1

    logger.info(f"set_device_mesh: {device_mesh}")


def get_device_mesh():
    global TORCH_DEVICE_MESH
    return TORCH_DEVICE_MESH


def device_mesh_world_size(device_mesh=None):
    if device_mesh is None:
        device_mesh = get_device_mesh()

    if device_mesh is None:
        return None

    if isinstance(device_mesh, TorchMockDeviceMesh):
        device_mesh_shape = get_device_mesh().shape
    elif isinstance(device_mesh, DeviceMesh):
        device_mesh_shape = tuple(get_device_mesh().mesh.shape)

    return functools.reduce(operator.mul, device_mesh_shape)


def device_mesh_rank(device_mesh, dim):
    if device_mesh is None:
        device_mesh = get_device_mesh()

    if isinstance(device_mesh, TorchMockDeviceMesh):
        global_rank = torch.distributed.distributed_c10d.get_rank()
        device_mesh_shape = get_device_mesh().shape
        world_size = functools.reduce(operator.mul, device_mesh_shape)
        rank_coords = (numpy.arange(world_size).reshape(
            *device_mesh_shape) == global_rank).nonzero()
        rank = rank_coords[dim].item()
    elif isinstance(device_mesh, DeviceMesh):
        rank = device_mesh.get_coordinate_on_dim(dim)
    else:
        raise RuntimeError("DeviceMesh not support or not initialize")

    return rank



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import logging

log_level_str = os.environ.get("EASYDIST_LOGLEVEL", 'INFO').upper()
log_level = logging.getLevelName(log_level_str)

dump_dir = os.environ.get("EASYDIST_DUMP_PATH", './md_dump')
enable_compile_cache = os.environ.get("ENABLE_COMPILE_CACHE", "False").upper() in ["1", "TRUE"]
compile_cache_dir = os.environ.get("EASYDIST_COMPILE_CACHE_PATH", './md_compiled')

easydist_device = os.environ.get("EASYDIST_DEVICE", "cuda")

# MetaSPMD Annotation

use_hint = os.environ.get("EASYDIST_USE_HINT") == "1"
extend_space = False

# Solver

liveness_only_input = False
max_seconds_same_incumbent = float('inf')

all_to_all_punish_factor = 3.

enable_graph_coarsen = os.environ.get("ENABLE_GRAPH_COARSEN", "True").upper() in ["1", "TRUE"]
coarsen_level = int(os.environ.get("COARSEN_LEVEL", "1"))

# runtime

use_contiguous_buffer = False



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import logging

import jax
from mpi4py import MPI

from .api import easydist_shard, get_opt_strategy, set_device_mesh
from .sharding_interpreter import EDJaxShardingAnn
from .bridge import jax2md_bridge

__all__ = [
    "easydist_shard", "get_opt_strategy", "set_device_mesh", "EDJaxShardingAnn", "jax2md_bridge"
]

logger = logging.getLogger(__name__)


def is_jax_distributed_initialized():
    return jax._src.distributed.global_state.client is not None


def easydist_setup_jax(device, allow_tf32):
    os.environ["NVIDIA_TF32_OVERRIDE"] = "1" if allow_tf32 else "0"
    jax.config.update('jax_platforms', device)

    # setup distributed
    comm = MPI.COMM_WORLD
    size, rank = comm.Get_size(), comm.Get_rank()

    if not is_jax_distributed_initialized():

        jax.distributed.initialize(coordinator_address="localhost:19705",
                                   num_processes=size,
                                   process_id=rank,
                                   local_device_ids=rank)

        logging.info(
            f"[Rank {rank}], Global Devices: {jax.device_count()}, Local Devices: {jax.local_device_count()}"
        )



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch
from torch.fx.node import _get_qualified_name


def md_embedding(weight, indices, padding_idx=-1, scale_grad_by_freq=False, sparse=False):
    if int(torch.max(indices).item()) >= weight.shape[0]:
        raise RuntimeError("embedding indice overflow")
    return torch.ops.aten.embedding.default(weight, indices, padding_idx, scale_grad_by_freq,
                                            sparse)


def fix_embedding(fx_module: torch.fx.GraphModule, recover=False):

    for node in fx_module.graph.nodes:
        if node.op == 'call_function':
            if "torch.ops.aten.embedding.default" in _get_qualified_name(node.target):
                node.target = md_embedding

            if recover and "md_embedding" in _get_qualified_name(node.target):
                node.target = torch.ops.aten.embedding.default

    fx_module.recompile()

    return fx_module



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import logging

from . import platform
import easydist.config as mdconfig


def easydist_setup(backend, device="cpu", allow_tf32=True):
    mdconfig.easydist_device = device

    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                        datefmt='%m/%d %H:%M:%S',
                        level=mdconfig.log_level)

    if backend == "jax":
        from .jax import easydist_setup_jax
        easydist_setup_jax(device, allow_tf32)

        logging.getLogger("jax._src").setLevel(logging.INFO)
    elif backend == "torch":
        from .torch import easydist_setup_torch
        easydist_setup_torch(device, allow_tf32)

        logging.getLogger("torch._subclasses.fake_tensor").setLevel(logging.INFO)

    platform.init_backend(backend)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
from typing import List

from easydist import platform
from easydist.metashard.halo import HaloInfo


class ShardDim:
    """
    ``ShardDim`` is used to store sharding information for ``ShardAnnotation``.

    Within ``ShardAnnotation``, each dimension of the tensor list is assigned a ``ShardDim``. 
    A value of ``ShardDim(0)`` means that the dimension cannot be sharded. 
    All dimensions with a value of ``ShardDim(i)`` can be sharded together.

    ShardDim can also carry information about halo and chunk.

    - ``halo`` means that padding is needed for each shard.
    - ``chunk`` means that we need firstly chunk the data along the dimension,
      then shard for each chunk to get the sharded data. similar to block-cyclic.

      [1, 1, 2, 2] -(shard)-> [1, 1] | [2, 2] when chunk = 1
      [1, 1, 2, 2] -(chunk)-> [1, 1] and [2, 2] -(shard)-> [1, 2] | [1, 2]

    """

    def __init__(self, shard_dim_id: int, chunk: int = 1) -> None:
        self.shard_dim_id = shard_dim_id
        self.chunk = chunk

        self.halo: HaloInfo = None

    def __str__(self) -> str:
        if self.shard_dim_id == 0:
            return "NoShardDim"
        else:
            content = str(self.shard_dim_id)
            if self.chunk > 1:
                content += f", chunk={self.chunk})"
            if self.halo:
                content += f", halo={self.halo})"
            return f"ShardDim({content})"

    def set_halo(self, halo_: HaloInfo):
        self.halo = halo_

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def get_noshard_dim():
        return ShardDim(0)

    @staticmethod
    def get_shard_dim(shard_dim_id):
        return ShardDim(shard_dim_id)


class ShardAnnotation:
    """
    ``ShardAnnotation`` used to describe the sharding strategy space for a ``MetaOp``.
    
    For example:
    - matmul: ShardAnnotation([[ShardDim(1), ShardDim(2)], [ShardDim(2), ShardDim(3)]])
    - relu: ShardAnnotation([[ShardDim(1), ShardDim(2)], [ShardDim(1), ShardDim(2)]])
    - layernorm: ShardAnnotation([[ShardDim(1), ShardDim(2), NoShardDim]])
    """

    def __init__(self, annotation: List[List[ShardDim]]) -> None:
        self.annotation = annotation

    @staticmethod
    def init_from_input_args(input_args):
        return ShardAnnotation([[ShardDim.get_noshard_dim()] * len(t.shape) for t in input_args
                                if isinstance(t, platform.Tensor)])

    def inject_haloinfo(self, haloinfo: HaloInfo, shard_dim_idx: int):
        """set haloinfo for each ShardDim(shard_dim_idx) in ShardAnnotation"""
        if haloinfo is None:
            return
        for anno in self.annotation:
            for shard_dim in anno:
                if shard_dim.shard_dim_id == shard_dim_idx:
                    shard_dim.set_halo(haloinfo)

    def get_max_shard_dim_id(self) -> int:
        max_shard_dim_id = 0
        for anno in self.annotation:
            for dim_anno in anno:
                max_shard_dim_id = max(max_shard_dim_id, dim_anno.shard_dim_id)
        return max_shard_dim_id

    def clear_shard_dim(self, max_shard_dim_id):
        new_anno = copy.deepcopy(self)
        for anno in new_anno.annotation:
            for dim_idx in range(len(anno)):
                if anno[dim_idx].shard_dim_id > max_shard_dim_id:
                    anno[dim_idx] = ShardDim.get_noshard_dim()
        return new_anno

    def __str__(self) -> str:
        return f"ShardAnnotation({self.annotation.__str__()})"

    def __repr__(self) -> str:
        return f"ShardAnnotation({self.annotation.__repr__()})"

    def __getitem__(self, idx):
        return self.annotation[idx]

    def __setitem__(self, idx, shard_dim):
        self.annotation[idx] = shard_dim

    def __len__(self) -> int:
        return self.annotation.__len__()

    def __add__(self, other):
        return ShardAnnotation(self.annotation + other.annotation)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import setuptools
from importlib.machinery import SourceFileLoader

version = (
    SourceFileLoader("easydist.version", os.path.join(
        "easydist", "version.py")).load_module().VERSION
)


def is_comment_or_empty(line):
    stripped = line.strip()
    return stripped == "" or stripped.startswith("#")


def remove_comments_and_empty_lines(lines):
    return [line for line in lines if not is_comment_or_empty(line)]


def get_core_requirements():
    with open(os.path.join("requirements", "core-requirements.txt")) as f:
        core_requirements = remove_comments_and_empty_lines(
            f.read().splitlines())
    return core_requirements


def get_long_description():
    with open("README.md", "r") as fh:
        long_description = fh.read()
    return long_description


setuptools.setup(
    name="pai-easydist",
    version=version,
    author="Shenggan Cheng",
    author_email="shenggan.c@u.nus.edu",
    description="Efficient Automatic Training System for Super-Large Models",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/alibaba/easydist",
    packages=setuptools.find_packages(),
    install_requires=get_core_requirements(),
    extras_require={
        "torch": [
            "torch",
            "torchvision",
        ],
        "jax": [
            "jax[cuda11_pip]",
            "flax",
        ]
    },
)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging

import torch
import torch.utils._pytree as pytree
from torch.distributed._tensor import DTensor, Replicate
from torch.fx.node import Node, _get_qualified_name, map_arg

from easydist.torch.device_mesh import device_mesh_rank, get_device_mesh
from easydist.utils.testing import MockDeviceMesh

logger = logging.getLogger(__name__)


def materialize(x, device):
    if isinstance(x, torch.Tensor):
        if x.dtype == torch.bool:
            return torch.rand(x.size(), dtype=torch.float, device=device) > 0.5
        elif torch.is_floating_point(x):
            return torch.rand(x.size(), dtype=x.dtype, device=device)
        else:
            return torch.randint(high=8, size=x.size(), dtype=x.dtype, device=device)
    return x


def redist_tensor_func(input_tensor, specs):
    if isinstance(input_tensor, DTensor) and input_tensor.size() != torch.Size([0]):
        device_mesh = get_device_mesh()
        if specs != input_tensor._spec.placements:
            return input_tensor.redistribute(device_mesh, specs).contiguous()
    return input_tensor.contiguous()


def insert_spmd_head(body):
    return ["from torch.distributed._tensor import Shard, Replicate\n", *body]


@torch.no_grad()
def to_dtensor(input_tensor, placements=None, global_size=None):
    if isinstance(input_tensor, torch.Tensor) and not isinstance(input_tensor, DTensor):
        device_mesh = get_device_mesh()
        if placements is None:
            placements = [Replicate()] * device_mesh.ndim
        if global_size is None:
            return DTensor.from_local(input_tensor, device_mesh, placements)
        else:
            return DTensor(input_tensor, device_mesh, placements, size=global_size)
    return input_tensor


def fix_in_gragh_tensor(fx_module: torch.fx.GraphModule, sharding_strategy):

    for node in fx_module.graph.nodes:
        if node.op == 'get_attr':
            with fx_module.graph.inserting_after(node):
                to_dtensor_node = fx_module.graph.call_function(to_dtensor, args=(node, ))

                node.replace_all_uses_with(to_dtensor_node)

                to_dtensor_node.update_arg(0, node)

        create_ops = [
            "torch.ops.aten.scalar_tensor.", "torch.ops.aten.ones.", "torch.ops.aten.empty.",
            "torch.ops.aten.arange.", "torch.ops.aten.zeros."
        ]

        if node.op == 'call_function':
            for op in create_ops:
                if op in _get_qualified_name(node.target):
                    strategy, global_shape = None, None
                    if node.name in sharding_strategy and len(sharding_strategy[node.name]) > 0:
                        strategy = sharding_strategy[node.name][0]
                        # modify the arg of creat ops
                        # (CAUTION!!) device_mesh maybe mock now,
                        # but the local shape need to be consistent of runtime
                        global_shape = node.args[0]
                        local_shape = list(global_shape)
                        device_mesh = get_device_mesh()
                        if isinstance(device_mesh, MockDeviceMesh):
                            logger.warning("maybe wrong shape in MockDeviceMesh for create ops.")
                        for idx, placement in enumerate(strategy):
                            if placement.is_shard():
                                shard_dim = placement.dim
                                split_size, pad_idx = divmod(global_shape[shard_dim],
                                                             device_mesh.size(idx))
                                local_shape[shard_dim] = split_size
                                if device_mesh_rank(device_mesh, idx) < pad_idx:
                                    local_shape[shard_dim] = split_size + 1

                        node.update_arg(0, local_shape)

                    with fx_module.graph.inserting_after(node):
                        to_dtensor_node = fx_module.graph.call_function(to_dtensor,
                                                                        args=(node, strategy,
                                                                              global_shape))

                        node.replace_all_uses_with(to_dtensor_node)

                        to_dtensor_node.update_arg(0, node)

    fx_module.recompile()

    return fx_module


def replace_subsequence_use(node, arg_, redist_node):
    users_node = list(node.users.keys())
    node_next = node.next

    def maybe_replace_node(n: Node) -> Node:
        if n == arg_:
            return redist_node
        else:
            return n

    while node_next.name != "":
        if node_next in users_node:
            new_args = map_arg(node_next.args, maybe_replace_node)
            new_kwargs = map_arg(node_next.kwargs, maybe_replace_node)
            assert isinstance(new_args, tuple)
            assert isinstance(new_kwargs, dict)
            node_next.args = new_args
            node_next.kwargs = new_kwargs
        node_next = node_next.next


def sharding_transform(fx_module: torch.fx.GraphModule, sharding_strategy):
    for node in fx_module.graph.nodes:
        if node.op == 'call_function':
            if node.name in sharding_strategy:
                node_strategy = sharding_strategy[node.name]

                # skip for create ops
                ops_name = _get_qualified_name(node.target)
                if ops_name in [
                        "torch.ops.aten.empty.memory_format", "torch.ops.aten.zeros.default"
                ]:
                    continue

                node_args_flatten = pytree.tree_flatten(node.args)[0]
                invars = [arg for arg in node_args_flatten if isinstance(arg, Node)]
                assert (len(invars) == len(node_strategy))

                for arg_, arg_strategy in zip(invars, node_strategy):
                    with fx_module.graph.inserting_before(node):
                        redist_node = fx_module.graph.call_function(redist_tensor_func,
                                                                    args=(arg_, arg_strategy))

                        # FIXME: may introduce redundancy communication, update_arg for all subsequence use
                        node.replace_input_with(arg_, redist_node)
                        #replace_subsequence_use(node, arg_, redist_node)

    fx_module.graph.on_generate_code(lambda _: insert_spmd_head)

    fx_module.recompile()

    # (fix) %_tensor_constant0 : [#users=1] = get_attr[target=_tensor_constant0]
    fx_module = fix_in_gragh_tensor(fx_module, sharding_strategy)

    return fx_module



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split('.'))



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tvm
import numpy
import torch.utils._pytree as pytree


def add(a, b):
    return tvm.nd.array(numpy.add(a.numpy(), b.numpy()))


def equal(a, b):
    return numpy.array_equal(a.numpy(), b.numpy())


def zeros_like(input_):
    return tvm.nd.array(numpy.zeros_like(input_.numpy()))


def min(a, b):
    return tvm.nd.array(numpy.minimum(a.numpy(), b.numpy()))


def max(a, b):
    return tvm.nd.array(numpy.maximum(a.numpy(), b.numpy()))


def allclose(a, b):
    return numpy.allclose(a.numpy(), b.numpy(), rtol=5e-3, atol=5e-03)


def concatenate(tensors, dim=0):
    return tvm.nd.array(numpy.concatenate([t.numpy() for t in tensors], axis=dim))


def chunk(input, chunks, dim=0):
    return [tvm.nd.array(i) for i in numpy.array_split(input.numpy(), chunks, axis=dim)]


def narrow(input, dim, start, length):
    indices = numpy.asarray(range(start, start + length))
    return tvm.nd.array(numpy.take(input.numpy(), indices, axis=dim))


Tensor = tvm.nd.NDArray

tree_flatten = pytree.tree_flatten
tree_unflatten = pytree.tree_unflatten


def clone(input_):
    return tvm.nd.array(numpy.copy(input_.numpy()))


from_numpy = tvm.nd.array



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import logging
import importlib

logger = logging.getLogger(__name__)

EASYDIST_BACKEND = None

__all__ = [
    "add", "equal", "zeros_like", "min", "max", "allclose", "concatenate", "chunk", "narrow",
    "Tensor", "tree_flatten", "tree_unflatten", "clone", "from_numpy"
]


def backend_valid(_backend):
    return _backend in {"torch", "jax", "tvm"}


def init_backend(backend="torch"):
    assert backend_valid(backend)
    global EASYDIST_BACKEND
    EASYDIST_BACKEND = backend
    modules = importlib.import_module("." + backend, __name__)
    for val in __all__:
        exec("globals()['%s'] = modules.%s" % (val, val))
    logger.info(f"========= EasyDist init with backend {backend}. =========")


def get_backend():
    global EASYDIST_BACKEND
    return EASYDIST_BACKEND


for val in __all__:
    exec("globals()['%s'] = None" % val)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import jax

from easydist.metashard.metair import MetaGraph, MetaNode, MetaVar


def jax2md_bridge(jaxpr: jax.core.Jaxpr, sharding_info, meta_info) -> MetaGraph:

    meta_graph = MetaGraph(jaxpr)

    meta_var_map = {}

    for in_var in jaxpr.invars + jaxpr.constvars:
        meta_var = MetaVar(name=in_var.__str__(),
                           shape=meta_info[in_var.__str__()]["shape"],
                           dtype=meta_info[in_var.__str__()]["dtype"])
        meta_var_map[in_var.__str__()] = meta_var

        node_sharding_info = None
        onelevel_key = "placeholder"
        twolevel_key = str([(meta_var.shape, meta_var.dtype)])
        if onelevel_key in sharding_info:
            node_sharding_info = sharding_info[onelevel_key].get(twolevel_key, None)

        meta_node = MetaNode(name=in_var.__str__(),
                             op_name="placeholder",
                             invars=[],
                             outvars=[meta_var],
                             sharding_info=node_sharding_info,
                             is_placeholder=True)
        meta_graph.add_node(meta_node)
        meta_graph.add_input(meta_node)

    for eqn in jaxpr.eqns:
        subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
        onelevel_key = eqn.primitive.__str__()
        if eqn.primitive.__str__() == "custom_jvp_call":
            onelevel_key += "[" + subfuns[0].f.args[0].eqns[0].params['name'] + "]"
        if eqn.primitive.__str__() in ["xla_call", "pjit"]:
            onelevel_key += "[" + eqn.params['name'] + "]"

        node_sharding_info = None
        if onelevel_key in sharding_info:
            abstract_list = []
            for var in eqn.invars:
                if var.__str__() in meta_info:
                    abstract_list.append(
                        (meta_info[var.__str__()]["shape"], meta_info[var.__str__()]["dtype"]))
                elif type(var) is jax.core.Literal:
                    abstract_list.append(var.val)
                else:
                    abstract_list.append(var)
            twolevel_key = str(abstract_list) + str(bind_params)
            if twolevel_key in sharding_info[onelevel_key]:
                node_sharding_info = sharding_info[onelevel_key][twolevel_key]

        outvars_ = []
        for var in eqn.outvars:
            name = var.__str__()
            outvar = MetaVar(name=name,
                             shape=meta_info[name]["shape"],
                             dtype=meta_info[name]["dtype"])
            outvars_.append(outvar)
            meta_var_map[name] = outvar

        # NOTE, use last outvars for name of MetaNode in Jax
        meta_node = MetaNode(name=name,
                             op_name=onelevel_key,
                             invars=[
                                 meta_var_map[var.__str__()] for var in eqn.invars
                                 if isinstance(var, jax.core.Var)
                             ],
                             outvars=outvars_,
                             sharding_info=node_sharding_info)
        meta_graph.add_node(meta_node)

    for var in jaxpr.outvars:
        meta_graph.add_output(meta_var_map[var.__str__()])

    return meta_graph
